/******************************************************************************
*******************************************************************************
*******************************************************************************

    libferris
    Copyright (C) 2001 Ben Martin

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

    For more details see the COPYING file in the root directory of this
    distribution.

    $Id: FerrisEAGeneratorPlugin.cpp,v 1.4 2007/08/29 21:30:09 ben Exp $

*******************************************************************************
*******************************************************************************
******************************************************************************/

#include <FerrisEAGeneratorPlugin.hh>
#include <FerrisEAGeneratorPlugin_private.hh>

namespace Ferris
{
    
    StaticEAGenFactorys_t& getStaticLinkedEAGenFactorys()
    {
        static StaticEAGenFactorys_t ret;
        return ret;
    }
    
    bool RegisterEAGeneratorModule(
        const fh_matcher& ma,
        const std::string& implname,
        const std::string& shortname,
        bool isDynamic,
        bool hasState,
        AttributeCreator::CreatePri_t CreatePri )
    {
        StaticEAGenFactorys_t& fac = getStaticLinkedEAGenFactorys();

//         std::cerr << "RegisterEAGeneratorModule() shortname:" << shortname
//                   << " isDyn:" << isDynamic
//                   << " hasState:" << hasState
//                   << " pri:" << CreatePri
//                   << std::endl;
        
        fac.push_back(
            new StaticGModuleMatchedEAGeneratorFactory( ma,
                                                        implname,
                                                        shortname,
                                                        isDynamic,
                                                        hasState,
                                                        CreatePri ));
    }

    FERRISEXP_API void
    AppendAllStaticEAGeneratorFactories_Stateless(
        Context::s_StatelessEAGenFactorys_t& SL )
    {
        StaticEAGenFactorys_t& fac = getStaticLinkedEAGenFactorys();
        StaticEAGenFactorys_t::iterator end = fac.end();
        
        for( StaticEAGenFactorys_t::iterator fi = fac.begin(); fi != end; ++fi )
        {
            fh_StaticMatchedEAGeneratorFactory f = *fi;

            if( !f->hasState() )
                SL.push_back( f );
        }
    }
    
    FERRISEXP_API bool
    AppendAllStaticEAGeneratorFactories_Statefull(
        Context::m_StatefullEAGenFactorys_t& SF )
    {
        bool HaveDynamicAttributes = false;
        
        StaticEAGenFactorys_t& fac = getStaticLinkedEAGenFactorys();
        StaticEAGenFactorys_t::iterator end = fac.end();
        
        for( StaticEAGenFactorys_t::iterator fi = fac.begin(); fi != end; ++fi )
        {
            fh_StaticMatchedEAGeneratorFactory f = *fi;

            if( f->hasState() )
                SF.push_back( f->clone() );

            HaveDynamicAttributes |= f->isDynamic();
        }

        return HaveDynamicAttributes;
    }
    
    

    
//     EAGenFactorys_t& AppendAllStaticEAGeneratorFactories( EAGenFactorys_t& ret )
//     {
//         StaticEAGenFactorys_t& fac = getStaticLinkedEAGenFactorys();
//         StaticEAGenFactorys_t::iterator end = fac.end();
        
//         for( StaticEAGenFactorys_t::iterator fi = fac.begin(); fi != end; ++fi )
//         {
//             fh_StaticMatchedEAGeneratorFactory f = *fi;

//             if( f->hasState() )
//             {
//                 ret.push_back( f->clone() );
//             }
//             else
//             {
//                 ret.push_back( f );
//             }
//         }
//         return ret;
//     }
    
    
    
};
