/******************************************************************************
*******************************************************************************
*******************************************************************************

    libferris
    Copyright (C) 2001 Ben Martin

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

    For more details see the COPYING file in the root directory of this
    distribution.

    $Id: FullTextQuery.hh,v 1.4 2006/06/09 13:44:28 ben Exp $

*******************************************************************************
*******************************************************************************
******************************************************************************/

#ifndef _ALREADY_INCLUDED_FERRIS_FULLTEXTQUERY_H_
#define _ALREADY_INCLUDED_FERRIS_FULLTEXTQUERY_H_

#include <Ferris/HiddenSymbolSupport.hh>
#include <Ferris/TypeDecl.hh>
#include <Ferris/FullTextIndexer.hh>

#include <string>

namespace Ferris
{
    /**
     * Support for full text indexing and querys
     */
    namespace FullTextIndex 
    {
        FERRISEXP_API enum QueryMode {
            QUERYMODE_USERPREF = 1<<1,
            QUERYMODE_BOOLEAN  = 1<<2,
            QUERYMODE_RANKED   = 1<<3,
            QUERYMODE_XAPIAN   = 1<<4,
            QUERYMODE_TSEARCH2 = 1<<5,
            QUERYMODE_EXTERNAL = 1<<6,
            QUERYMODE_BEAGLE   = 1<<7
        };
        std::string makeFullTextQueryFilterString( const std::string& qs,
                                                   QueryMode qm,
                                                   fh_idx idx );
        
        /**
         * Execute the given query in the mode selected against either the given
         * index or the default index if no index is given
         */
        FERRISEXP_API fh_context ExecuteQuery( const std::string& s,
                                               QueryMode qm = QUERYMODE_USERPREF,
                                               fh_idx idx = 0 );

        class FullTextQuery;
        FERRIS_SMARTPTR( FullTextQuery, fh_ftquery );

        namespace Factory 
        {
            /**
             * Used for more advanced queries than ExecuteQuery() can provide.
             * first call here to create a Query object and then you can setup
             * ranked query cutoffs and other tweaks that are set to default
             * values if ExecuteQuery() is used.
             */
            FERRISEXP_API fh_ftquery makeFullTextQuery( const std::string& s,
                                                        QueryMode qm = QUERYMODE_USERPREF,
                                                        fh_idx idx = 0 );
        };
        
        class FERRISEXP_API FullTextQuery
            :
            public Handlable
        {
            typedef FullTextQuery _Self;
            typedef Handlable     _Base;

        protected:
            
            fh_idx m_idx;
            int    m_accumulatorsMaxSize;
            int    m_resultSetMaxSize;
            std::string m_queryString;

            FullTextQuery( const std::string& s, fh_idx idx );
            
        public:

            virtual ~FullTextQuery();

            /********************************************************************************/
            /********************************************************************************/
            /*** the following have effect for all query types ******************************/
            /********************************************************************************/
            /********************************************************************************/

            /**
             * set the index to use for future executions
             */
            void setIndex( fh_idx idx );
            
            /**
             * set the query string
             */
            void setQuery( const std::string& s );

            /**
             * run the query
             */
            virtual fh_context execute() = 0;
            
            /********************************************************************************/
            /********************************************************************************/
            /*** the following are only effective for RANKED queries ************************/
            /********************************************************************************/
            /********************************************************************************/

            /**
             * max Number of candidate documents to calculate for
             */
            void setAccumulatorsMaxSize( int v );

            /**
             * only take the top v documents from the ranked results
             */
            void setResultSetMaxSize( int v );
        };
    };
};
#endif
