/******************************************************************************
*******************************************************************************
*******************************************************************************

    libferris
    Copyright (C) 2001 Ben Martin

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

    For more details see the COPYING file in the root directory of this
    distribution.

    $Id: MetadataServer.cpp,v 1.2 2007/11/23 21:30:14 ben Exp $

*******************************************************************************
*******************************************************************************
******************************************************************************/

#include "MetadataServer_private.hh"

#include "DBusGlue/ferris-internal-metadata-broker-client-glue.h"

#include <Ferris/Ferris.hh>
#include <iostream>

#include <dbus-c++/glib-integration.h>


using namespace std;

namespace Ferris
{
    static const char* DBUS_BROKER = "net.sf.witme.Libferris.Metadata.Broker";
    static const char* DBUS_SERVER_PATH = "/net/sf/witme/libferris/Metadata";


    namespace
    {
        class MetadataBroker
            :
            public net::sf::witme::Libferris::Metadata::Broker,
            public DBus::IntrospectableProxy,
            public DBus::ObjectProxy
        {
        public:

            DBus::Dispatcher* dispatcher;

            bool is_error;
            std::vector< ::DBus::Byte > value;
            string errormsg;
            bool running;
            
            MetadataBroker( DBus::Connection& connection,
                              const char* path, const char* name )
                :
                DBus::ObjectProxy( connection, path, name )
                , dispatcher( DBus::default_dispatcher )
                {}

            void run()
                {
                    LG_MDSERV_D << "run() running:" << running << endl;
//                    dispatcher->enter();
                    while( running )
                        g_main_iteration( true );
                    
                }
        
        

            virtual void asyncGetResult( const ::DBus::Int32& reqid,
                                         const ::DBus::String& earl,
                                         const ::DBus::String& rdn,
                                         const std::vector< ::DBus::Byte >& value )
                {
                    LG_MDSERV_D << "asyncGetResult() req:" << reqid
                                << " earl:" << earl
                                << " rdn:" << rdn
                                << " value:" << value
                                << endl;
                    
                    this->is_error = 0;
                    this->value = value;
                    this->errormsg = "";

                    running = false;
//                    dispatcher->leave();
                }

            virtual void asyncGetFailed( const ::DBus::Int32& reqid,
                                         const ::DBus::String& earl,
                                         const ::DBus::Int32& error,
                                         const ::DBus::String& ename,
                                         const ::DBus::String& edesc )
                {
                    LG_MDSERV_D << "asyncGetFailed() req:" << reqid
                                << " earl:" << earl
                                << " error:" << error
                                << endl;

                    this->is_error = 1;
                    this->value.clear();
                    this->errormsg = tostr(error);
                
                    running = false;
//                    dispatcher->leave();
                }
        
            virtual void asyncPutCommitted( const ::DBus::Int32& reqid,
                                            const ::DBus::String& earl,
                                            const ::DBus::String& rdn )
                {}
        
            virtual void asyncPutFailed( const ::DBus::Int32& reqid,
                                         const ::DBus::String& earl,
                                         const ::DBus::Int32& errno,
                                         const ::DBus::String& ename,
                                         const ::DBus::String& edesc )
                {}
        
        
        };

        struct NestedDBusDispatcher
        {
            DBus::Dispatcher* restore;
            DBus::BusDispatcher d;
            NestedDBusDispatcher()
                {
                    restore = DBus::default_dispatcher;
                    DBus::default_dispatcher = &d;
                }
            ~NestedDBusDispatcher()
                {
                    DBus::default_dispatcher = restore;
                }
        };
    };

    
    DBus::Connection getSessionBus()
    {
        if( !DBus::default_dispatcher )
        {
            LG_MDSERV_D << "setting a glib dbus++ dispatcher..." << endl;
            cerr << "setting a glib dbus++ dispatcher..." << endl;
            static DBus::Glib::BusDispatcher* dispatcher = 0;
            dispatcher = new DBus::Glib::BusDispatcher();
            DBus::default_dispatcher = dispatcher;
            dispatcher->attach(NULL);
        }
        static DBus::Connection conn = DBus::Connection::SessionBus();
        return conn;
    }

    std::string syncMetadataServerGet( const std::string& earl,
                                       const std::string& rdn,
                                       DBus::Connection conn )
    {
        LG_MDSERV_D << "syncMetadataServerGet()" << endl;
        
        LG_MDSERV_D << "done setting a dispatcher..." << endl;
//        NestedDBusDispatcher d;
        Ferris::MetadataBroker client(conn, DBUS_SERVER_PATH, DBUS_BROKER );

        LG_MDSERV_D << "Calling asyncGet() earl:" << earl << " rdn:" << rdn << endl;
        client.running = true;

        int rc = client.asyncGet( earl, rdn );
        LG_MDSERV_D << "Calling asyncGet() rc:" << rc << endl;

        LG_MDSERV_D << "Calling run()" << endl;
        client.run();

        LG_MDSERV_D << "Have a reply is_error:" << client.is_error << endl;
        if( client.is_error )
        {
            stringstream ss;
            ss << "error:" << client.errormsg;
            Throw_DBusException( ss.str(), 0 );
        }

        string ret( client.value.begin(), client.value.end() );
        return ret;
    }

    void syncMetadataServerPut( const std::string& earl,
                                const std::string& rdn,
                                const std::string& value,
                                DBus::Connection con )
    {
        LG_MDSERV_D << "FIXME!" << endl;
        Throw_DBusException( "not implemented", 0 );
    }

    /************************************************************/
    /************************************************************/
    /************************************************************/

    static guint glib_idle_timeout_ID = 0;
    static gint glib_idle_timeout_func(gpointer data)
    {
        bool* running = (bool*)data;
        (*running) = false;
    }

    void resetup_glib_idle_to_close_callback_function( int interval, bool* running )
    {
        if( glib_idle_timeout_ID  )
            g_source_remove( glib_idle_timeout_ID );
        glib_idle_timeout_ID = g_timeout_add_seconds(
            interval, GSourceFunc(glib_idle_timeout_func), 0 );
    }
    
    /************************************************************/
    /************************************************************/
    /************************************************************/

    stringset_t& getImageMetadataAttributes()
    {
        static stringset_t* ret = 0;
        if( !ret )
        {
            ret = new stringset_t;
            string alist = "depth-per-color,depth,gamma,has-alpha,aspect-ratio,width,height,rgba-32bpp,megapixels";
            Util::parseCommaSeperatedList( alist, *ret );
        }
        return *ret;
    }

    bool isImageMetadataAttribute( const std::string& rdn )
    {
        return( getImageMetadataAttributes().end() == getImageMetadataAttributes().find( rdn ) );
    }
    
    s_ImageEAGeneratorsExtensionToShortName_t& getOutOfProcess_ImageEAGeneratorsExtensionToShortName()
    {
        static s_ImageEAGeneratorsExtensionToShortName_t* ret = 0;
        if( !ret )
        {
            string shortNamesToUseString = "xine,imlib2,jpeg,magick,png";

            ret = new s_ImageEAGeneratorsExtensionToShortName_t();
            s_ImageEAGeneratorsExtensionToShortName_t& all = getImageEAGeneratorsExtensionToShortName();
            stringset_t shortNamesToUse;
            Util::parseCommaSeperatedList( shortNamesToUseString, shortNamesToUse );

            s_ImageEAGeneratorsExtensionToShortName_t::iterator iter = all.begin();
            s_ImageEAGeneratorsExtensionToShortName_t::iterator    e = all.end();

            for( ; iter != e; ++iter )
            {
                string shortname = iter->second.first;
                if( shortNamesToUse.end() != shortNamesToUse.find( shortname ) )
                {
                    ret->insert( *iter );
                }
            }
            
        }
        return *ret;
    }

    const stringset_t&
    getOutOfProcess_EAGeneratorsStaticFactoriesShortNamesToUse()
    {
        string shortNamesToUseString = "xine,kde3_metadata";
        static stringset_t shortNamesToUse;
        Util::parseCommaSeperatedList( shortNamesToUseString, shortNamesToUse );
        cerr << "shortNamesToUse.sz:" << shortNamesToUse.size() << endl;
        return shortNamesToUse;
    }
    
    
    StaticEAGenFactorys_t&
    getOutOfProcess_EAGeneratorsStaticFactories( StaticEAGenFactorys_t& ret )
    {
        const stringset_t& shortNamesToUse = getOutOfProcess_EAGeneratorsStaticFactoriesShortNamesToUse();

        StaticEAGenFactorys_t& fac = getStaticLinkedEAGenFactorys();
        StaticEAGenFactorys_t::iterator end = fac.end();
        for( StaticEAGenFactorys_t::iterator fi = fac.begin(); fi != end; ++fi )
        {
            fh_StaticMatchedEAGeneratorFactory f = *fi;
            string sn = f->getShortName();
            cerr << "sn:" << sn << endl;
            if( shortNamesToUse.end() != shortNamesToUse.find( sn ) )
            {
                ret.push_back( f );
            }
        }

        cerr << "MetadataBrokerDispatchInformation() selected factories sz:"
             << ret.size() << endl;

        return ret;
    }

    bool isOutOfProcessMetadataAttribute( const std::string& rdn )
    {
        static stringset_t col;
        if( col.empty() )
        {
            string alist = "";

            // Xine EA
            alist += "width,height,bitrate,can-seek,video-channel-count,video-stream-count,video-bitrate,";
            alist += "video-bitrate-fourcc,xine-can-play-video,audio-channel-count,audio-bitrate,";
            alist += "audio-bitrate-fourcc,xine-can-play-audio,has-video,has-audio,title,comment,artist,";
            alist += "genre,album,year,";
            
            Util::parseCommaSeperatedList( alist, col );
        }

        return col.end() != col.find( rdn );
    }
    
    
}

