/******************************************************************************
*******************************************************************************
*******************************************************************************

    redland RDF private support functions.  

    Copyright (C) 2003 Ben Martin

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

    For more details see the COPYING file in the root directory of this
    distribution.

    $Id: Redland.cpp,v 1.21 2007/12/27 21:30:11 ben Exp $

*******************************************************************************
*******************************************************************************
******************************************************************************/

#include <boost/archive/text_oarchive.hpp>
#include <boost/archive/text_iarchive.hpp>
#include <boost/archive/binary_oarchive.hpp>
#include <boost/archive/binary_iarchive.hpp>
#include <boost/serialization/list.hpp>
#include <boost/serialization/set.hpp>
#include <boost/serialization/map.hpp>
#include <boost/serialization/split_member.hpp>

#include <Ferris/Ferris.hh>
#include <Ferris/Redland_private.hh>
#include <Ferris/Trimming.hh>
#include <Ferris/FerrisBoost.hh>
#include <Ferris/Regex.hh>

#include <rdf_node.h>
#include <rdf_query.h>

// #undef LG_RDF_D
// #define LG_RDF_D if( 0 ) cerr

namespace boost {
    namespace serialization {

        template<class Archive>
        void serialize(Archive & ar, ::Ferris::fh_regex& e, const unsigned int version)
        {
            ar & *(GetImpl(e));
        }

    } // namespace serialization
} // namespace boost

namespace Ferris
{
    using namespace std;
    
    
    namespace RDF 
    {
        const std::string RDF_FERRIS_BASE = "http://witme.sf.net/libferris-core/0.1";

        
        typedef std::list< fh_node > nodelist_t;
        
        static librdf_world* W()
        {
            return getWorld()->getRAW();
        }
        
        static std::string qyn( bool v )
        {
            if( v )
                return "'yes'";
            return "'no'";
        }

        std::string convertstr( char* v )
        {
            if( v )
            {
                std::string ret = v;
                free( v );
                return ret;
            }
            return "";
        }
        
        struct FERRISEXP_DLLLOCAL ImplWorld
        {
            static fh_uri URI( librdf_uri* u, fh_world w = 0 )
                {
                    return new ::Ferris::RDF::URI( u );
                }

            static fh_node CopyNode( librdf_node* n, fh_world w = 0 )
                {
                    fh_node ret = new ::Ferris::RDF::Node( librdf_new_node_from_node( n ), true );
                    return ret;
                }
            static fh_node Node( librdf_node* n, fh_world w = 0 )
                {
                    if( !n )
                        return 0;
                    
                    fh_node ret = new ::Ferris::RDF::Node( n, true );
                    return ret;
                }
        };
        
        World::World()
            :
            m_world( 0 )
        {
            LG_RDF_D << "World::World()" << endl;
            
            m_world = librdf_new_world();
            librdf_world_open( m_world );
        }
        
        World::~World()
        {
//            LG_RDF_D << "World::~World()" << endl;
        }

        librdf_world*
        World::getRAW()
        {
            return m_world;
        }
        
        fh_world
        getWorld()
        {
            static fh_world w = new World();
//            w->AddRef();
            return w;
            
//             static World w;
//             w.AddRef();
//             static fh_world wdummy = &w;
//             return wdummy;
        }

        fh_model
        getDefaultFerrisModel()
        {
            fh_world w = getWorld();
            static fh_model m = 0;

            if( m )
                return m;
            
            //
            // Allow user customization of which storage to use for their RDF
            //
            bool usingCustomStore = false;
            try
            {
                fh_context rdfdbc = Resolve( "~/.ferris/rdfdb" );
                std::string storage_name = getStrSubCtx( rdfdbc, "ferris-storage-name", "" );
                std::string db_name      = getStrSubCtx( rdfdbc, "ferris-db-name", "~/.ferris/rdfdb/myrdf" );
                std::string db_options   = getStrSubCtx( rdfdbc, "ferris-db-options", "" );

                if( !storage_name.empty() )
                {
//                     cerr << "storage_name:" << storage_name
//                          << " db_name:" << db_name
//                          << " db_options:" << db_options
//                          << endl;
                
                    usingCustomStore = !storage_name.empty();
                    m = Model::ObtainDB( storage_name, db_name, db_options );
                    return m;
                }
            }
            catch( exception& e )
            {
                if( usingCustomStore )
                    throw;
            }

            fh_context dbc = Shell::acquireContext( "~/.ferris/rdfdb" );
            m = Model::ObtainDB( dbc, "myrdf", "~/.ferris/rdfdb" );
//            m->AddRef();
            return m;
        }

        /********************************************************************************/
        /********************************************************************************/
        /********************************************************************************/

        Storage::Storage( const std::string& storage_name,
                          const std::string& name,
                          const std::string& exopts )
            :
            m_storage( 0 )
        {
            m_storage = librdf_new_storage( W(),
                                            (char*)storage_name.c_str(),
                                            (char*)name.c_str(),
                                            (char*)(exopts.empty() ? 0 : exopts.c_str()) );
//             LG_RDF_D << "Storage::Storage() format:" << storage_name
//                  << " m_storage:" << toVoid( m_storage ) << endl;
            
        }

        fh_storage
        Storage::CreateStorage( const std::string& storage_name,
                                const std::string& name,
                                const std::string& options_string )
        {
            return new Storage( storage_name, name, options_string );
        }
        fh_storage
        Storage::CreateDBStorage( const std::string& name ,
                                  const std::string& dir  ,
                                  bool writable )
        {
            fh_stringstream ss;
            ss << "hash-type='bdb',dir='" << dir
//               << "',write='" << (writable ? "yes" : "no")
               << "'";
            return new Storage( "hashes", name, tostr(ss) );
        }
        
        Storage::~Storage()
        {
            if( m_storage )
            {
                librdf_free_storage( m_storage );
            }
        }

        librdf_storage*
        Storage::getRAW()
        {
            return m_storage;
        }
        
        fh_model
        Storage::CreateModel( const std::string& exopts )
        {
            fh_model m = new Model( this, exopts );
            return m;
        }
        
        
        /********************************************************************************/
        /********************************************************************************/
        /********************************************************************************/

        Model::Model( fh_storage s, const std::string& exopts )
            :
            m_storage( s ),
            m_model( 0 ),
            m_rdfxml_url( "" )
        {
            m_model = librdf_new_model( W(), s->getRAW(),
                                        (char*)(exopts.empty() ? 0 : exopts.c_str()) );
            if( !m_model )
            {
                fh_stringstream ss;
                ss << "Model::Model can not open model from storage." << endl;
                ss << " rdf.world:" << toVoid( W() )
                   << " storage:" << s->getRAW();
                ss << " exopts:" << exopts << endl;
                // FIXME: throw!
                cerr << tostr(ss) << endl;
                BackTrace();
                throw;
            }
        }

        Model::Model( fh_storage s, librdf_model* m )
            :
            m_storage( s ),
            m_model( m ),
            m_rdfxml_url( "" )
        {
        }

        fh_model
        Model::Load( fh_context c )
        {
            fh_storage storage = 0;
            fh_model   ret     = 0;
            std::string ex = getStrAttr( c, "name-extension", "" );

            LG_RDF_D << "Model::Load() ext:" << ex << " c:" << c->getURL() << endl;
            
            if( !cmp_nocase( ex, "xml" ) || !cmp_nocase( ex, "rdf" ) )
            {
                LG_RDF_D << "Model::Load() rdf/xml for url:" << c->getURL() << endl;
                
                storage            = Storage::CreateStorage( "memory", c->getDirPath(), "" );
                fh_model   model   = storage->CreateModel();

                model->m_rdfxml_url = c->getURL();

                fh_parser  parser  = new Parser();
                std::string rdfxml_content = getStrAttr( c, "content", "", true, true );
//                LG_RDF_D << "rdfxml_content -->:" << rdfxml_content << ":<--" << endl;
                parser->ParseIntoModel( model, rdfxml_content, new URI( c->getURL() ) );
                
                return model;
            }
            else if( !cmp_nocase( ex, "db" ))
            {
                LG_RDF_D << "Model::Load() name:" << c->getDirName()
                         << " path:" << c->getDirPath()
                         << " url:" << c->getURL()
                         << endl;
                
                std::string name = c->getDirName();
                std::string path = c->getParent()->getDirPath();
                
                PostfixTrimmer ptrimmer;

                ptrimmer.push_back( "-contexts.db" );
                ptrimmer.push_back( "-p2so.db" );
                ptrimmer.push_back( "-po2s.db" );
                ptrimmer.push_back( "-so2p.db" );
                ptrimmer.push_back( "-sp2o.db" );
                name = ptrimmer( name );

                LG_RDF_D << "Model::Load() name:" << name
                         << " path:" << path << endl;
                
                storage = Storage::CreateDBStorage( name, path, true );
                ret     = storage->CreateModel();

//                 LG_RDF_D << "Model::Load(dumping) name:" << c->getDirName()
//                      << " path:" << c->getDirPath()
//                      << " url:" << c->getURL()
//                      << endl;
//                 ret->write( Factory::fcerr() );
//                 Factory::fcerr() << flush;
                
            }
            
            return ret;
        }

        fh_model
        Model::Load( const std::string& rdfxml )
        {
            fh_storage storage = Storage::CreateStorage( "memory", "", "" );
            fh_model   model   = storage->CreateModel();

            fh_parser  parser  = new Parser();
            parser->ParseIntoModel( model, rdfxml, new URI( "file:///dev/null" ) );
            return model;
        }

        void
        Model::MergeRDFXML( const std::string& rdfxml, fh_uri base_uri )
        {
            fh_parser  parser  = new Parser();
            parser->ParseIntoModel( this, rdfxml, base_uri );
        }
        
        fh_model
        Model::ObtainDB( fh_context c,
                         const std::string& name,
                         const std::string& dbenv_name,
                         bool writable,
                         bool indexContexts,
                         bool indexPredicates,
                         bool createAsNew )
        {
            std::string storage_name = "hashes";

            fh_stringstream ss;
            ss << "hash-type='bdb',"
               << "dir='" << c->getDirPath() << "',"
               << "write=" << qyn( writable ) << ","
               << "contexts=" << qyn( indexContexts ) << ","
               << "index-predicates=" << qyn( indexPredicates ) << ","
               << "new=" << qyn( createAsNew );
            if( !dbenv_name.empty() )
            {
                string p = dbenv_name;
                p = Util::replace_all( p, "~", Shell::getHomeDirPath() + "/" );
                ss << "db-environment-dir='" << p << "'";
            }

//            cerr << "Model::ObtainDB(2) ss:" << tostr(ss) << endl;
            
            fh_storage storage = Storage::CreateStorage( storage_name, name, tostr(ss) );
            LG_RDF_D << "Model::ObtainDB(1) ss:" << tostr(ss)
                     << " storage:" << toVoid( storage->getRAW() )
                     << endl;
            fh_model ret = storage->CreateModel();
            LG_RDF_D << "Model::ObtainDB(2) ss:" << tostr(ss) << endl;
            return ret;
        }

        fh_model
        Model::ObtainDB( fh_context c,
                         const char* name,
                         const char* dbenv_name,
                         bool writable,
                         bool indexContexts,
                         bool indexPredicates,
                         bool createAsNew )
        {
            return ObtainDB( c, (string)name, (string)dbenv_name, writable, indexContexts, indexPredicates, createAsNew );
        }
        
        fh_model
        Model::ObtainDB( fh_context c,
                         const std::string& name,
                         bool writable,
                         bool indexContexts,
                         bool indexPredicates,
                         bool createAsNew )
        {
//             cerr << "Model::ObtainDB(ctx,name)" << endl;
//             cerr << " ctx:" << c->getURL() << endl;
//             cerr << " name:" << name << endl;
            return ObtainDB( c, name, "", writable, indexContexts, indexPredicates, createAsNew );
        }
        fh_model
        Model::ObtainDB( const std::string& path,
                         const std::string& name,
                         bool writable,
                         bool indexContexts,
                         bool indexPredicates,
                         bool createAsNew )
        {
//            cerr << "Model::ObtainDB(path,name)" << endl;
            fh_context c = Shell::acquireContext( path );
            return ObtainDB( c, name, writable, indexContexts, indexPredicates, createAsNew );
        }
        fh_model
        Model::ObtainDB( const std::string& storage_name,
                         const std::string& db_name,
                         const std::string& db_options )
        {
//             cerr << "Model::ObtainDB(storage_name,db,options)" << endl;
//             cerr << "storage_name:" << storage_name << endl;
//             cerr << "db_name:" << db_name << endl;
//             cerr << "db_options:" << db_options << endl;
            fh_storage storage = Storage::CreateStorage( storage_name, db_name, db_options );
            fh_model ret = storage->CreateModel();
            return ret;
        }
        
        fh_model
        Model::MemoryOnlyModel()
        {
            fh_storage storage = Storage::CreateStorage();
            fh_model ret = storage->CreateModel();
            return ret;
        }
        
        
        
        
        Model::~Model()
        {
            if( m_model )
                librdf_free_model( m_model );
        }

        librdf_model*
        Model::getRAW() const
        {
            return m_model;
        }
        
        fh_model
        Model::clone()
        {
            librdf_model* m = librdf_new_model_from_model( m_model );
            return new Model( m_storage, m );
        }

        void
        Model::sync()
        {
            librdf_model_sync( getRAW() );

            if( !m_rdfxml_url.empty() )
            {
                fh_ofstream oss( m_rdfxml_url );
                write( oss );
            }
        }
        
        
        int
        Model::size()
        {
            return librdf_model_size( m_model );
        }

        void
        Model::insert( fh_statement st )
        {
            return addStatement( st );
        }
        
        void
        Model::insert( StatementIterator iter, StatementIterator end )
        {
            for( ; iter != end ; ++iter )
            {
                insert( *iter );
            }
        }

        void
        Model::insert( fh_node subj, fh_node pred, fh_node obj )
        {
            insert( new Statement( subj, pred, obj ) );
        }

        void
        Model::insert( fh_model m )
        {
            if( librdf_model_add_submodel( getRAW(), m->getRAW() ))
            {
                LG_RDF_ER << "inserting a submodel failed!" << endl;
            }
        }
        
        
        StatementIterator
        Model::begin()
        {
            librdf_stream* s = librdf_model_as_stream( getRAW() );
            return StatementIterator( s );
        }
        
        StatementIterator
        Model::end()
        {
            return StatementIterator();
        }
        
        
        
        void
        Model::addStatement( fh_statement st )
        {
            LG_RDF_D << "Model::addStatement() st     :" << toVoid( st ) << endl;
            LG_RDF_D << "Model::addStatement() st.raw :" << toVoid( st->getRAW() ) << endl;
            LG_RDF_D << "Model::addStatement() m_model:" << toVoid( m_model ) << endl;
            LG_RDF_D << "Model::addStatement() already_contains:"
                     << librdf_model_contains_statement( getRAW(), st->getRAW() )
                     << endl;

            int rc = librdf_model_add_statement( getRAW(), st->getRAW() );
            LG_RDF_D << "Model::addStatement() rc:" << rc << endl;
            
            if( st->getSubject() )
            {
                LG_RDF_D << "Model::addStatement() subj:" << st->getSubject()->toString() << endl;
            }
            if( st->getPredicate() )
            {
                LG_RDF_D << "Model::addStatement() pred:" << st->getPredicate()->toString() << endl;
            }
            if( st->getObject() )
            {
                LG_RDF_D << "Model::addStatement() obj:" << st->getObject()->toString() << endl;
            }
            
            LG_RDF_D << "Model::addStatement(end)" << endl;
        }

        fh_ostream
        Model::write( fh_ostream oss,
                      const std::string& serializerName,
                      const std::string& mimeType,
                      fh_uri type_uri )
        {
            librdf_serializer* serializer = 0;
            FILE* tfile = tmpfile();

            serializer = librdf_new_serializer(
                W(),
                serializerName.empty() ? 0 : serializerName.c_str(),
                mimeType.empty()       ? 0 : mimeType.c_str(),
                type_uri               ? type_uri->getRAW() : 0
                );
            
//            int rc = librdf_serializer_serialize_model( serializer, tfile, 0, m_model );
            int rc = librdf_serializer_serialize_model_to_file_handle( serializer, tfile, 0, m_model );
            if( rc != 0 )
            {
                // error
                LG_RDF_ER << "Error serializing the RDF model!" << endl;
            }
            rewind( tfile );

            const size_t bufsz = 1024;
            char buf[ bufsz + 1 ];
            while( size_t bread = fread( buf, 1, bufsz, tfile ))
            {
                oss.write( buf, bread );
            }

            if( !feof( tfile ) )
            {
                // error
                LG_RDF_ER << "Failed to read in all the data from temporary storage!" << endl;
            }
            fclose( tfile );

            oss << flush;
            return oss;
        }

        class FERRISEXP_API StatementIteratorQRHolder
            :
            public StatementIterator
        {
            typedef StatementIteratorQRHolder _Self;
            friend class Model;

            librdf_query *query;
            
        public:

            typedef std::forward_iterator_tag iterator_category;
            typedef fh_statement value_type;
            typedef long         difference_type;
            typedef fh_statement pointer;
            typedef fh_statement reference;

            explicit StatementIteratorQRHolder( librdf_query *query, librdf_stream* stream )
                :
                StatementIterator( stream ),
                query( query )
                {}
            
            StatementIteratorQRHolder()             {}
            virtual ~StatementIteratorQRHolder()
                {
                    if( query )
                        librdf_free_query( query );
                }
            reference operator*();
        };
        

        StatementIterator
        Model::find( const std::string& query,
                     const std::string& queryTypeName,
                     fh_uri queryTypeURI )
        {
//             librdf_stream* stream = librdf_model_query_string(
//                 getRAW(),
//                 queryTypeName.c_str(),
//                 queryTypeURI ? queryTypeURI->getRAW() : 0,
//                 (const unsigned char*)query.c_str() );

//             return StatementIterator( stream );

            librdf_query* q = librdf_new_query( W(),
                                                queryTypeName.c_str(),
                                                queryTypeURI ? queryTypeURI->getRAW() : 0,
                                                (const unsigned char*)query.c_str(),
                                                0 );
            librdf_query_results* qr = librdf_model_query_execute( getRAW(), q );
            librdf_stream* stream = librdf_query_results_as_stream( qr );
//             cerr << " q:" << q
//                  << " qr:" << qr
//                  << " stream:" << stream
//                  << endl;
            
            return StatementIteratorQRHolder( q, stream );
        }

        /********************************************************************************/
        /********************************************************************************/
        /********************************************************************************/

        void BindingsIterator::binding::setup( librdf_query_results* qr )
        {
            if( m_bindingNames.empty() )
            {
                int bc = librdf_query_results_get_bindings_count( qr );
                for( int i=0; i<bc; ++i )
                {
                    const char* bname = librdf_query_results_get_binding_name( qr, i );
                    m_bindingNames.push_back( bname );

//                     librdf_node* raw_n = librdf_query_results_get_binding_value( qr, i );
//                     cerr << "binding::setup() NAMES i:" << i << " bname:" << bname
//                          << " raw_n:" << raw_n 
//                          << endl;
                }
            }
            
            int bc = librdf_query_results_get_bindings_count( qr );

            d.clear();
            for( int i=0; i<bc; ++i )
            {
                librdf_node* raw_n = librdf_query_results_get_binding_value( qr, i );
                fh_node n = new Node( raw_n, true );
                d.push_back( n );
//                cerr << "binding::setup() i:" << i << " raw_n:" << raw_n << " n:" << GetImpl(n) <<  endl;
            }
        }

        int BindingsIterator::binding::size() const
        {
            return d.size();
        }
        
        const BindingsIterator::binding::m_bindingNames_t&
        BindingsIterator::binding::getNames() const
        {
            return m_bindingNames;
        }
        
        fh_node
        BindingsIterator::binding::getByName( const std::string& n ) const
        {
            m_bindingNames_t::const_iterator ni = find( m_bindingNames.begin(), m_bindingNames.end(), n );
            if( ni == m_bindingNames.end() )
                return 0;
            m_bindingNames_t::const_iterator b = m_bindingNames.begin();
            int id = distance( b, ni );
            return getByID( id );
        }
        fh_node
        BindingsIterator::binding::getByID( int id ) const
        {
//            cerr << "binding::getByID id:" << id << " d.sz:" << d.size() << " ret:" << d[ id ] << endl;
            return d[ id ];
        }
        std::string
        BindingsIterator::binding::getName( int n ) const
        {
            m_bindingNames_t::const_iterator ni = m_bindingNames.begin();
            advance( ni, n );
            return *ni;
        }
        
        BindingsIterator::BindingsIterator( librdf_query_results* iter )
            :
            m_iter( new librdf_query_results_holder( iter ) )
        {
            if( 0 != librdf_query_results_finished( iter ) )
            {
                __free();
            }
        }
        
        
        librdf_query_results*
        BindingsIterator::__iter()
        {
            return m_iter->m_iter;
        }
        
        
        BindingsIterator::reference
        BindingsIterator::operator*()
        {
            if( m_iter )
            {
                librdf_query_results* i = __iter();
                m_binding.setup( i );
            }
            return m_binding;
        }

        void
        BindingsIterator::__free()
        {
            m_iter = 0;
        }
        
        void
        BindingsIterator::__advance()
        {
            if( m_iter )
            {
                librdf_query_results_next( __iter() );
                if( 0 != librdf_query_results_finished( __iter() ) )
                    __free();
            }
        }

        bool operator==(const BindingsIterator& __x, const BindingsIterator& __y)
        {
            return __x.m_iter == __y.m_iter;
        }
        bool operator!=(const BindingsIterator& __x, const BindingsIterator& __y)
        {
            return !(__x == __y);
        }
        
        
        BindingsIterator
        Model::findBindings( const std::string& query,
                             const std::string& queryTypeName,
                             fh_uri queryTypeURI )
        {
//             rasqal_query *rq = rasqal_new_query( queryTypeName.c_str(),
//                                                  queryTypeURI ? queryTypeURI->getRAW() : 0 );
//             if(rasqal_query_prepare(rq, (const unsigned char*)query.c_str(), 0 ))
//             {
//                 cerr << "failed to prepare query" << endl;
//                 return BindingsIterator();
//             }
//             rasqal_query_results *results;
//             if(!(results=rasqal_query_execute(rq))) {
//                 cerr << "failed to get results" << endl;
//                 return BindingsIterator();
//             }
//             if(!rasqal_query_results_is_bindings(results)) {
//             {
//                 cerr << "expecting bindings and query not returning them!" << endl;
//                 return BindingsIterator();
//             }
            
//             cerr << "Model::findBindings() 4" << endl;
//             return BindingsIterator( qr );


            
            
//            cerr << "Model::findBindings() 1" << endl;
            librdf_query* q = librdf_new_query( W(),
                                                queryTypeName.c_str(),
                                                queryTypeURI ? queryTypeURI->getRAW() : 0,
                                                (const unsigned char*)query.c_str(),
                                                0 );
//            cerr << "Model::findBindings() 2" << endl;
            librdf_query_results* qr = librdf_model_query_execute( getRAW(), q );
//            cerr << "Model::findBindings() 3" << endl;

            if( !librdf_query_results_is_bindings ( qr ) )
            {
                cerr << "expecting bindings and query not returning them!" << endl;
                return BindingsIterator();
            }
            
//            cerr << "Model::findBindings() 4" << endl;
            return BindingsIterator( qr );
        }

        /********************************************************************************/
        /********************************************************************************/
        /********************************************************************************/
        
        StatementIterator
        Model::findStatements( fh_statement partial_statement )
        {
            librdf_stream* stream = librdf_model_find_statements(
                getRAW(), partial_statement->getRAW() );
            return StatementIterator( stream );
        }


        ObjectIterator
        Model::findSubjects(   fh_node pred, fh_node obj )
        {
            LG_RDF_D << "findSubjects() pred:" << toVoid( pred )
                     << " obj:" << toVoid( obj )
                     << endl;
            
            librdf_iterator* iter = librdf_model_get_sources(
                getRAW(), pred->getRAW(), obj->getRAW() );
            return ObjectIterator( iter );
        }
            
        ObjectIterator
        Model::findPredicates( fh_node subj, fh_node obj )
        {
            LG_RDF_D << "findPredicates() subj:" << toVoid( subj )
                     << " obj:" << toVoid( obj ) <<endl;
            librdf_iterator* iter = librdf_model_get_arcs(
                getRAW(), subj->getRAW(), obj->getRAW() );
            return ObjectIterator( iter );
        }
            
        ObjectIterator
        Model::findObjects( fh_node subj, fh_node pred )
        {
            LG_RDF_D << "findObjects() subj:" << toVoid( subj )
                     << " pred:" << toVoid( pred )
                     << endl;
            librdf_iterator* iter = librdf_model_get_targets(
                getRAW(), subj->getRAW(), pred->getRAW() );
            return ObjectIterator( iter );
        }
        
        fh_node
        Model::getSubject(     fh_node pred, fh_node obj )
        {
            librdf_node* n = librdf_model_get_source( getRAW(), pred->getRAW(), obj->getRAW() );
            return ImplWorld::Node( n );
        }
        
        fh_node
        Model::getPredicate(   fh_node subj, fh_node obj )
        {
            librdf_node* n = librdf_model_get_arc( getRAW(), subj->getRAW(), obj->getRAW() );
            return ImplWorld::Node( n );
        }
        
        fh_node
        Model::getObject(      fh_node subj, fh_node pred )
        {
            librdf_node* n = librdf_model_get_target( getRAW(), subj->getRAW(), pred->getRAW() );
            return ImplWorld::Node( n );
        }

        ObjectIterator
        Model::getArcsIn(  fh_node n )
        {
//             LG_RDF_D "getArcsIn() n:" << toVoid( n )
//                  << endl;
            librdf_iterator* iter = librdf_model_get_arcs_in( getRAW(), n->getRAW() );
            return ObjectIterator( iter );
        }
        
        ObjectIterator
        Model::getArcsOut( fh_node n )
        {
//             LG_RDF_D << "Model::getArcsOut() this:" << toVoid( this )
//                  << " n:" << toVoid( n )
//                  << " n.impl:" << toVoid( GetImpl( n ))
//                  << " raw:" << toVoid( getRAW() )
//                  << " n.raw:" << toVoid( n->getRAW() )
//                  << endl;
            librdf_iterator* iter = librdf_model_get_arcs_out( getRAW(), n->getRAW() );
            return ObjectIterator( iter );
        }
        
        bool
        Model::hasArcIn(   fh_node n, fh_node p )
        {
            int rc = librdf_model_has_arc_in( getRAW(), n->getRAW(), p->getRAW() );
            return rc;
        }
        
        bool
        Model::hasArcOut(  fh_node n, fh_node p )
        {
            int rc = librdf_model_has_arc_out( getRAW(), n->getRAW(), p->getRAW() );
            return rc;
        }
        
        bool
        Model::contains( fh_statement st )
        {
            bool ret = librdf_model_contains_statement( getRAW(), st->getRAW() );
            return ret;
        }
        
        void
        Model::erase( fh_statement st )
        {
            LG_RDF_D << "Model::erase(st) pred:" << st->getPredicate()->toString()
                 << " obj:" << st->getObject()->toString()
                 << endl;
            librdf_model_remove_statement( getRAW(), st->getRAW() );
        }

        void
        Model::erase( fh_node subj, fh_node pred, fh_node obj )
        {
            erase( new Statement( subj, pred, obj ) );
        }

        void
        Model::erase( fh_node subj, fh_node pred )
        {
            ObjectIterator end   = ObjectIterator();
            ObjectIterator oiter = findObjects( subj, pred );
            list< fh_statement > stack;
            for( ; oiter != end; ++oiter )
            {
                fh_statement st = new Statement( subj, pred, *oiter );
                stack.push_back( st );
            }
            LG_RDF_D << "Model::erase(s,p) found count:" << stack.size() << endl;

            for( list< fh_statement >::iterator iter = stack.begin();
                 iter != stack.end(); ++iter )
            {
                erase( *iter );
            }
        }

        void
        Model::erase( fh_node subj, fh_node pred, ObjectIterator iter )
        {
            nodelist_t nl;
//            copy( iter, ObjectIterator(), back_inserter(nl) );
            for( ; iter != ObjectIterator(); ++iter )
            {
                LG_RDF_D << "Model::erase(objiter) obj:" << (*iter)->toString() << endl;
                nl.push_back( *iter );
            }
            
            LG_RDF_D << "Model::erase( object iter ) nl.size:" << nl.size() << endl;
            
            for( nodelist_t::iterator ni = nl.begin(); ni != nl.end(); )
            {
                fh_node n = *ni;
                ++ni;
                erase( subj, pred, n );
            }
        }
        
        void
        Model::erase( fh_model m )
        {
            if( librdf_model_remove_submodel( getRAW(), m->getRAW() ))
            {
                LG_RDF_D << "removing a submodel failed!" << endl;
            }
        }

        void
        Model::eraseTransitive( fh_node subj )
        {
            StatementList_t sl;
            getTransitiveTree( subj, sl );
            for( StatementList_t::iterator si = sl.begin(); si!=sl.end(); ++si )
            {
                LG_RDF_D << "Erasing subj:" << (*si)->getObject()->toString()
                         << " pred:" << (*si)->getPredicate()->toString()
                         << " obj:" << (*si)->getObject()->toString()
                         << endl;
                erase( *si );
            }
        }
        
        void
        Model::set( fh_node subj, fh_node pred, fh_node obj )
        {
            LG_RDF_D << "Model::set() obj:" << obj->toString() << endl;
            erase( subj, pred );
            insert( subj, pred, obj );
        }
        

        StatementList_t&
        Model::getTransitiveTree( fh_node subj, StatementList_t& ret )
        {
            ObjectIterator end   = ObjectIterator();
            ObjectIterator piter = getArcsOut( subj );
            for( ; piter != end; ++piter )
            {
                ObjectIterator oiter = findObjects( subj, *piter );
                for( ; oiter != end; ++oiter )
                {
                    fh_statement st = new Statement( subj, *piter, *oiter );
                    ret.push_back( st );
                    getTransitiveTree( *oiter, ret );
                }
            }
            return ret;
        }


        void
        Model::RDFMerge( fh_node source, fh_node target )
        {
            fh_node subj = source;
    
            ObjectIterator end;
            ObjectIterator piter = getArcsOut( subj );
            for( ; piter != end; ++piter )
            {
                ObjectIterator oiter = findObjects( subj, *piter );
                for( ; oiter != end; ++oiter )
                {
                    fh_statement st = new Statement( target, *piter, *oiter );

                    ObjectIterator io = findObjects( target, *piter );
                    if( io == end )
                    {
                        insert( st );
                    }
                }
            }
        }
        
        
        /********************************************************************************/
        /********************************************************************************/
        /********************************************************************************/

        Parser::Parser( const std::string& parser_name,
                        const std::string& mime_type,
                        fh_uri type_uri )
            :
            m_parser( 0 )
        {
            m_parser = librdf_new_parser( W(),
                                          parser_name.empty() ? 0 : parser_name.c_str(),
                                          mime_type.empty()   ? 0 : mime_type.c_str(),
                                          type_uri ? type_uri->getRAW() : 0 );
            if( !m_parser )
            {
                LG_RDF_ER << "Failed to create parser!" << endl;
            }
        }
        
            
        Parser::~Parser()
        {
            if( m_parser )
                librdf_free_parser( m_parser );
        }

        librdf_parser*
        Parser::getRAW()
        {
            return m_parser;
        }

        fh_model
        Parser::ParseIntoModel( fh_model m, fh_uri input_uri, fh_uri base_uri )
        {
            if( librdf_parser_parse_into_model( m_parser,
                                                input_uri->getRAW(),
                                                base_uri ? base_uri->getRAW() : 0,
                                                m->getRAW() ))
            {
                LG_RDF_ER << "Failed to parse RDF into model." << endl;
            }
            return m;
        }

        static int ParseIntoModel_error_fn(void *user_data, const char *msg, va_list arguments )
        {
            LG_RDF_ER << "Parse failure. msg:" << msg << endl;
            return 1;
        }
        
        fh_model
        Parser::ParseIntoModel( fh_model m, const std::string& s, fh_uri base_uri )
        {
//             cerr << "Parser::ParseIntoModel(string) s:" << s
//                  << " parser:" << m_parser
//                  << " base_uri:" << toVoid( base_uri )
//                  << " raw:" << m->getRAW()
//                  << endl;
            //void (*error_fn) (void *user_data, const char *message, va_list arguments));
//            librdf_parser_set_error(   m_parser, this, ParseIntoModel_error_fn );
//            librdf_parser_set_warning( m_parser, this, ParseIntoModel_error_fn );
            librdf_world_set_error( W(), this, ParseIntoModel_error_fn );
            LG_RDF_D << "Parser::ParseIntoModel(string)" << endl;
            fh_uri base_null = new URI( "file:///dev/null" );
            if( librdf_parser_parse_string_into_model( m_parser,
                                                       (const unsigned char*)s.c_str(),
                                                       base_uri
                                                       ? base_uri->getRAW()
                                                       : base_null->getRAW(),
                                                       m->getRAW() ))
            {
                LG_RDF_W << "Failed to parse RDF into model." << endl;
            }
            return m;
        }

        fh_model
        Parser::ParseIntoModel( fh_model m, fh_context c, fh_uri base_uri )
        {
            fh_istream iss = c->getIStream();
            return ParseIntoModel( m, iss, base_uri );
        }
        
        fh_model
        Parser::ParseIntoModel( fh_model m, fh_istream iss, fh_uri base_uri )
        {
            std::string s = StreamToString( iss );
            return ParseIntoModel( m, s, base_uri );
        }
        
        
        void
        Parser::setFeature( fh_uri feature, const std::string& s )
        {
            fh_node n = Node::CreateLiteral( s );
            
            if( librdf_parser_set_feature( getRAW(), feature->getRAW(),
                                           n->getRAW() ))
            {
                LG_RDF_D << "Failed to set librdf feature." << endl;
            }
        }
        
        void
        Parser::clearFeature( fh_uri feature )
        {
            fh_node n = Node::CreateLiteral( "" );
            if( librdf_parser_set_feature( getRAW(), feature->getRAW(),
                                           n->getRAW() ) )
            {
                LG_RDF_D << "Failed to set librdf feature." << endl;
            }
        }
        
        std::string
        Parser::getFeature( fh_uri feature )
        {
            fh_node n = ImplWorld::Node(
                librdf_parser_get_feature( getRAW(), feature->getRAW() ));
            return n->toString();
            
//             const char* v = librdf_parser_get_feature( getRAW(), feature->getRAW() );
//             std::string ret = v ? v : "";
//             return ret;
        }
        
        
        /********************************************************************************/
        /********************************************************************************/
        /********************************************************************************/

        Node::Node( librdf_node* n, bool takeOwnership )
            :
            m_node( n ),
            m_ownNode( takeOwnership )
        {
        }

        fh_node
        Node::CreateURI( const std::string& uri )
        {
            librdf_node* n = librdf_new_node_from_uri_string( W(),
                                                              (const unsigned char*)uri.c_str() );
            if( !n )
            {
                LG_RDF_W << "Node::CreateURI() failed!" << endl;
            }
            return new Node(  n, true );
        }
        
        fh_node
        Node::CreateLiteral( const std::string& s )
        {
            librdf_node* n = librdf_new_node_from_literal( W(),
                                                           (const unsigned char*)s.c_str(),
                                                           0, 0 );
            if( !n )
            {
                LG_RDF_W << "Node::CreateLiteral() failed!" << endl;
            }
            return new Node(  n, true );
        }

        fh_node
        Node::CreateLiteral( const std::wstring& s )
        {
            librdf_node* n = librdf_new_node_from_literal( W(),
                                                           (const unsigned char*)s.c_str(),
                                                           0, 0 );
            if( !n )
            {
                LG_RDF_W << "Node::CreateLiteral() failed!" << endl;
            }
            return new Node(  n, true );
        }

        fh_node
        Node::CreateBlankNode( const std::string& s )
        {
            librdf_node* n = librdf_new_node_from_blank_identifier(
                W(), s.empty() ? (unsigned char *)0 : (unsigned char *)s.c_str() );
            if( !n )
            {
                LG_RDF_W << "Node::CreateBNode() failed!" << endl;
            }
            return new Node(  n, true );
        }

        
        Node::~Node()
        {
            if( m_node && m_ownNode )
            {
                librdf_free_node( m_node );
                m_node = 0;
            }
        }
        
        librdf_node*
        Node::getRAW() const
        {
            return m_node;
        }

        fh_uri
        Node::getURI()
        {
            if( librdf_uri* u = librdf_node_get_uri( getRAW() ) )
                return ImplWorld::URI( librdf_new_uri_from_uri( u ) );
            return 0;
        }
        
        librdf_node_type
        Node::getType()
        {
            return librdf_node_get_type( getRAW() );
        }
        
        std::string
        Node::getLiteralValue()
        {
            char* v = (char*)librdf_node_get_literal_value( getRAW() );
            LG_RDF_D << "Node::getLiteralValue() v:" << v << endl;
            
            if( v )
                return v;
            return "(null)";
        }
        
        std::string
        Node::getLiteralLang()
        {
            char* v = (char*)librdf_node_get_literal_value_language( getRAW() );
            if( v )
                return v;
            return "en";
        }
        
        bool
        Node::isLiteralXML()
        {
            return librdf_node_get_literal_value_is_wf_xml( getRAW() );
        }
        
        fh_uri
        Node::getLiteralDatatype()
        {
            librdf_uri* u = librdf_node_get_literal_value_datatype_uri( getRAW() );
            if( u )
            {
                librdf_uri* myu = librdf_new_uri_from_uri( u );
                return ImplWorld::URI( u );
            }
            return 0;
        }
        
        int
        Node::getLIOrdinal()
        {
            return librdf_node_get_li_ordinal( getRAW() );
        }
        
        std::string
        Node::getBlankID()
        {
            char* v = (char*)librdf_node_get_blank_identifier ( getRAW() );
            return v ? v : "";
        }
        
        bool
        Node::isResource()
        {
            return librdf_node_is_resource ( getRAW() );
        }
        
        bool
        Node::isLiteral()
        {
            return librdf_node_is_literal ( getRAW() );
        }
        
        bool
        Node::isBlank()
        {
            return librdf_node_is_blank ( getRAW() );
        }
        
        std::string
        Node::toString()
        {
            if( isLiteral() )
                return getLiteralValue();
            
            char* v = (char*)librdf_node_to_string( getRAW() );
            return convertstr( v );
        }
        
//         std::string
//         Node::getDigest()
//         {
//             librdf_digest* d = librdf_node_get_digest ( getRAW () );
//             char* v = librdf_digest_to_string ( d );
//             return convertstr( v );
//         }
        
        std::string&
        Node::encode( std::string& ret )
        {
            size_t needed = librdf_node_encode( getRAW(), 0, 0 );
            ret.resize( needed );
            librdf_node_encode( getRAW(), (unsigned char*)ret.data(), needed );
            return ret;
        }
        
        
        bool operator==(const fh_node& __x, const fh_node& __y)
        {
            return *__x == *__y;
        }
        bool operator==(const Node& __x, const Node& __y)
        {
            return librdf_node_equals( __x.getRAW(), __y.getRAW() );
        }
        
        bool operator!=(const Node& __x, const Node& __y)
        {
            return !(__x == __y);
        }
        
        
        /********************************************************************************/
        /********************************************************************************/
        /********************************************************************************/

        Statement::Statement( )
            :
            m_ownStatement( true ),
            m_statement( 0 )
        {
            m_statement = librdf_new_statement( W() );
        }

        Statement::Statement( fh_node sub, fh_node pred, fh_node obj )
            :
            m_ownStatement( true ),
            m_statement( 0 )
        {
            LG_RDF_D << "Statement() w:" << toVoid( W() )
                 << " sub:" << toVoid( sub->getRAW() )
                 << " pred:" << toVoid( pred->getRAW() )
                 << " obj:" << toVoid( obj->getRAW() )
                 << endl;

            m_statement = librdf_new_statement_from_nodes(
                W(),
                librdf_new_node_from_node(  sub->getRAW() ),
                librdf_new_node_from_node( pred->getRAW() ),
                librdf_new_node_from_node(  obj->getRAW() )
                );
        }
        
        Statement::Statement( librdf_statement* st, bool takeOwnership )
            :
            m_statement( st ),
            m_ownStatement( takeOwnership )
        {
        }
        
        Statement::~Statement()
        {
            LG_RDF_D << "~Statement()" << endl;
            
            if( m_statement && m_ownStatement )
            {
                librdf_free_statement( m_statement );
            }
        }
            
        librdf_statement*
        Statement::getRAW() const
        {
            return m_statement;
        }

        void
        Statement::setSubject( fh_node n )
        {
            librdf_statement_set_subject( getRAW(),
                                          librdf_new_node_from_node(n->getRAW()) );
        }
        
        void
        Statement::setPredicate( fh_node n )
        {
            librdf_statement_set_predicate( getRAW(),
                                            librdf_new_node_from_node(n->getRAW()) );
        }
        
        void
        Statement::setObject( fh_node n )
        {
            librdf_statement_set_object( getRAW(),
                                         librdf_new_node_from_node(n->getRAW()) );
        }

        fh_node
        Statement::getSubject()
        {
            librdf_node* shn = librdf_statement_get_subject( getRAW() );
            return ImplWorld::CopyNode( shn );
        }
        
        fh_node
        Statement::getPredicate()
        {
            librdf_node* shn = librdf_statement_get_predicate( getRAW() );
            return ImplWorld::CopyNode( shn );
        }
        
        fh_node
        Statement::getObject()
        {
            librdf_node* shn = librdf_statement_get_object( getRAW() );
            return ImplWorld::CopyNode( shn );
        }

        void
        Statement::clear()
        {
            librdf_statement_clear( getRAW() );
        }
        
        bool
        Statement::isComplete()
        {
            return librdf_statement_is_complete( getRAW () );
        }
        
        std::string
        Statement::toString()
        {
            char* v = (char*)librdf_statement_to_string( getRAW () );
            return convertstr( v );
        }
        
        bool
        Statement::match( fh_statement partial_st )
        {
            int rc = librdf_statement_match( getRAW(), partial_st->getRAW() );
            return rc;
        }

        std::string
        Statement::toNTripleString( fh_model m )
        {
            return toXString( m, "ntriples" );
        }
        
        std::string
        Statement::toRDFXMLString( fh_model m )
        {
            return toXString( m, "rdfxml" );
        }

        class TransitiveRDFNodeCopy
        {
            fh_model m_outmodel;
            fh_model m_inmodel;
            fh_node m_inRootNode;
            fh_node m_newRootSubject;
            bool m_OverwriteExisting;
            bool m_createNewBNodes;
            
            typedef map< fh_node, fh_node, fh_node_compare > m_bnodeMapping_t;
            m_bnodeMapping_t m_bnodeMapping;
            
        public:
            TransitiveRDFNodeCopy( fh_model outmodel,
                                   fh_model inmodel,
                                   fh_node n,
                                   bool clone = true )
                :
                m_outmodel( outmodel ),
                m_inmodel( inmodel ),
                m_inRootNode( n ),
                m_newRootSubject( 0 ),
                m_OverwriteExisting( true ),
                m_createNewBNodes( false )
                {
                    if( !clone )
                    {
                        m_OverwriteExisting = false;
                        m_createNewBNodes = true;
                    }
                }

            void setNewRootSubject( fh_node n )
                {
                    m_newRootSubject = n;
                }
            void setOverwriteExisting( bool v )
                {
                    m_OverwriteExisting = v;
                }
            void setCreateNewBNodes( bool v )
                {
                    m_createNewBNodes = v;
                }

            void insertAllObjects( fh_node n, fh_node newRootSubject )
                {
                    LG_RDF_D << "TransitiveRDFNodeCopy::insertAllObjects(begin) "
                             << " n:" << n->toString()
                             << endl;
                    
                    ObjectIterator end   = ObjectIterator();
                    ObjectIterator piter = m_inmodel->getArcsOut( n );
                    for( ; piter != end; ++piter )
                    {
                        LG_RDF_D << "TransitiveRDFNodeCopy::insertAllObjects() "
                                 << " pred:" << (*piter)->toString()
                                 << endl;
                        
                        ObjectIterator oiter = m_inmodel->findObjects( n, *piter );
                        for( ; oiter != end; ++oiter )
                        {
                            LG_RDF_D << "TransitiveRDFNodeCopy::insertAllObjects() "
                                     << " obj:" << (*oiter)->toString()
                                     << endl;
                            
                            fh_statement st = new Statement( n, *piter, *oiter );
//                            if( !m_outmodel->contains( st ) )
                            {
                                LG_RDF_D << "TransitiveRDFNodeCopy::insertAllObjects(making) "
                                         << " m_createNewBNodes:" << m_createNewBNodes
                                         << " m_OverwriteExisting:" << m_OverwriteExisting
                                         << " newRootSubject:" << toVoid(newRootSubject)
                                         << " obj:" << (*oiter)->toString()
                                         << endl;
                                if( m_createNewBNodes )
                                {
                                    fh_node o = st->getObject();
                                    if( o->isBlank() )
                                    {
                                        fh_node b = Node::CreateBlankNode();
                                        st->setObject( b );
                                        m_bnodeMapping[ o ] = b;
                                    }
                                    fh_node s = st->getSubject();
                                    if( s->isBlank() )
                                    {
                                        fh_node b = m_bnodeMapping[ s ];
                                        st->setSubject( b );
                                    }
                                }
                                if( newRootSubject )
                                    st->setSubject( newRootSubject );

                                if( m_OverwriteExisting
                                    || !m_outmodel->getObject( st->getSubject(), st->getPredicate() ) )
                                {
                                    m_outmodel->insert( st );
                                }

                            }
                            LG_RDF_D << "TransitiveRDFNodeCopy::insertAllObjects(recurse) "
                                     << " object-becoming-node:" << (*oiter)->toString()
                                     << endl;
                            insertAllObjects( *oiter, 0 );
                        }
                    }

                    LG_RDF_D << "TransitiveRDFNodeCopy::insertAllObjects(end) "
                             << " n:" << n->toString()
                             << endl;
                }
            void operator()()
                {
                    insertAllObjects( m_inRootNode, m_newRootSubject );
                }
        };
        
//         typedef map< fh_node, fh_node, fh_node_compare > bnodeMapping_t;
//         bnodeMapping_t bnodeMapping;
//         static void
//         insertAllObjects( fh_model outmodel,
//                           fh_model inmodel,
//                           fh_node n,
//                           fh_node newRootSubject,
//                           bool dontSetExisting,
//                           bnodeMapping_t& bnodeMapping,
//                           bool createNewBNodes )
//         {
//             ObjectIterator end   = ObjectIterator();
//             ObjectIterator piter = inmodel->getArcsOut( n );
//             for( ; piter != end; ++piter )
//             {
//                 ObjectIterator oiter = inmodel->findObjects( n, *piter );
//                 for( ; oiter != end; ++oiter )
//                 {
//                     fh_statement st = new Statement( n, *piter, *oiter );
//                     if( !outmodel->contains( st ) )
//                     {
//                         if( createNewBNodes )
//                         {
//                             fh_node o = st->getObject();
//                             if( o->isBlank() )
//                             {
//                                 fh_node b = Node::CreateBlankNode();
//                                 st->setObject( b );
//                                 bnodeMapping[ o ] = b;
//                             }
//                             fh_node s = st->getSubject();
//                             if( s->isBlank() )
//                             {
//                                 fh_node b = bnodeMapping[ s ];
//                                 st->setSubject( b );
//                             }
//                         }
//                         if( newRootSubject )
//                             st->setSubject( newRootSubject );

//                         if( !dontSetExisting || !outmodel->getObject( st->getSubject(), st->getPredicate() ) )
//                             outmodel->insert( st );
                        
//                         insertAllObjects( outmodel, inmodel, *oiter, 0, dontSetExisting, bnodeMapping, createNewBNodes );
//                     }
//                 }
//             }
//         }
        /*
         * Use 'n' as a subject node and insert all of its transitive triples
         * into the output model 'outmodel'. Nodes are copied from 'inmodel'
         */
        static void
        insertAllObjects( fh_model outmodel, fh_model inmodel, fh_node n )
        {
            bool clone = true;
            TransitiveRDFNodeCopy trdf_copy( outmodel, inmodel, n, clone );
            trdf_copy();
        }

        
        std::string
        Statement::toXString( fh_model inmodel, const std::string& format )
        {
            fh_storage s = Storage::CreateStorage();
            fh_model m   = s->CreateModel();

            fh_stringstream ss;
            m->insert( getSubject(), getPredicate(), getObject() );
            insertAllObjects( m, inmodel, getObject() );
            m->write( ss, format );
            return tostr( ss );
        }
        
        bool operator==(const Statement& __x, const Statement& __y)
        {
            return librdf_statement_equals( __x.getRAW(), __y.getRAW() );
        }
        
        bool operator!=(const Statement& __x, const Statement& __y)
        {
            return !(__x == __y);
        }
        
        
        /********************************************************************************/
        /********************************************************************************/
        /********************************************************************************/
        
        RedlandStreamIterator::RedlandStreamIterator( librdf_stream* stream )
            :
            m_stream( 0 )
        {
            if( !librdf_stream_end( stream ))
                m_stream = new librdf_stream_holder( stream );
        }

        librdf_stream*
        RedlandStreamIterator::__stream()
        {
            return m_stream->m_stream;
        }
        
        void
        RedlandStreamIterator::__free()
        {
//            librdf_free_stream( m_stream );
            m_stream = 0;
        }
        
        
        void
        RedlandStreamIterator::__advance()
        {
            if( __stream() )
            {
                librdf_stream_next( __stream() );
                if( librdf_stream_end( __stream() ) )
                    __free();
            }
        }
        
        RedlandStreamIterator::RedlandStreamIterator()
            :
            m_stream( 0 )
        {
        }
        
        RedlandStreamIterator::~RedlandStreamIterator()
        {
            __free();
        }
        bool operator==(const RedlandStreamIterator& __x, const RedlandStreamIterator& __y)
        {
            return __x.m_stream == __y.m_stream;
        }
        bool operator!=(const RedlandStreamIterator& __x, const RedlandStreamIterator& __y)
        {
            return !(__x == __y);
        }

        /******************************/
        /******************************/
        /******************************/
        
        StatementIterator::reference
        StatementIterator::operator*()
        {
            reference ret = 0;
            
            if( m_stream )
            {
                librdf_statement* st = librdf_stream_get_object( __stream() );
                ret = new Statement( librdf_new_statement_from_statement( st ), true );
            }
            return ret;
        }


        /********************************************************************************/
        /********************************************************************************/
        /********************************************************************************/

        RedlandIterator::librdf_iterator_holder::librdf_iterator_holder( librdf_iterator* i )
            :
            m_iter( i )
        {
        }
        RedlandIterator::librdf_iterator_holder::~librdf_iterator_holder()
        {
            if( m_iter )
                librdf_free_iterator( m_iter );
        }

        /**********/
        
        RedlandIterator::RedlandIterator( librdf_iterator* iter )
            :
            m_iter( 0 )
        {
            if( !librdf_iterator_end( iter ))
            {
                m_iter = new librdf_iterator_holder( iter );
            }
            else
            {
                librdf_free_iterator( iter );
            }
        }

        librdf_iterator*
        RedlandIterator::__iter()
        {
            return m_iter->m_iter;
        }
        
        
        void
        RedlandIterator::__free()
        {
            m_iter = 0;
        }
        
        
        void
        RedlandIterator::__advance()
        {
            if( __iter() )
            {
                librdf_iterator_next( __iter() );
                if( librdf_iterator_end( __iter() ) )
                    __free();
            }
        }
        
        RedlandIterator::RedlandIterator()
            :
            m_iter( 0 )
        {
        }
        
        RedlandIterator::~RedlandIterator()
        {
            __free();
        }
        bool operator==(const RedlandIterator& __x, const RedlandIterator& __y)
        {
//             LG_RDF_D << "operator==( RIter ) x:" << toVoid( __x.m_iter )
//                  << " y:" << toVoid( __y.m_iter ) << endl;
            
            return  __x.m_iter == __y.m_iter;
        }
        bool operator!=(const RedlandIterator& __x, const RedlandIterator& __y)
        {
            return !(__x == __y);
        }

        /******************************/
        /******************************/
        /******************************/

        ObjectIterator::ObjectIterator( librdf_iterator* iter )
            :
            RedlandIterator( iter )
        {}
        
        ObjectIterator::ObjectIterator()
            :
            RedlandIterator( 0 )
        {
        }
        
        ObjectIterator::~ObjectIterator()
        {
        }
        
        ObjectIterator::reference
        ObjectIterator::operator*()
        {
            reference ret = 0;
            
            if( m_iter )
            {
                librdf_node* target = (librdf_node*)librdf_iterator_get_object( __iter() );
                ret = new Node( librdf_new_node_from_node( target ), true );
            }
            return ret;
        }
        
        /********************************************************************************/
        /********************************************************************************/
        /********************************************************************************/
        
        URI::URI( const std::string& uri )
            :
            m_uri( 0 )
        {
            m_uri = librdf_new_uri( W(),
                                    (const unsigned char*)uri.c_str() );
        }

        URI::URI( librdf_uri* u )
            :
            m_uri( 0 )
        {
            m_uri = u;
        }
            
        URI::~URI()
        {
        }
        
        librdf_uri*
        URI::getRAW() const
        {
            return m_uri;
        }
        

        std::string
        URI::toString()
        {
            char* v = (char*)librdf_uri_to_string( getRAW() );
            return convertstr( v );
        }
        
        std::string
        URI::getPath()
        {
            const char* v = librdf_uri_to_filename( getRAW() );
            return convertstr( (char*)v );
        }
        
        bool
        URI::isFile()
        {
            return librdf_uri_is_file_uri( getRAW() ) == 0;
        }
        
        bool operator==(const URI& __x, const URI& __y)
        {
            return librdf_uri_equals( __x.getRAW(), __y.getRAW() );
        }
        
        bool operator!=(const URI& __x, const URI& __y)
        {
            return !(__x == __y);
        }

    /********************************************************************************/
    /********************************************************************************/
    /********************************************************************************/

        namespace Wordnet
        {
            // word offset after the # in a word node [L#102037721cat]
            static const int WORD_NODE_TEXT_OFFSET = 9;

            string RDFSCHEMA = "http://www.dcc.uchile.cl/~agraves/wordnet/1.0/schema#";
            string WNPFX = "http://www.dcc.uchile.cl/~agraves/wordnet/1.0/";

            fh_node& hasWord()
            {
                static fh_node ret = Node::CreateURI( RDFSCHEMA + "hasWord" );
                return ret;
            }
            fh_node& hasSynSet()
            {
                static fh_node ret = Node::CreateURI( RDFSCHEMA + "hasSynSet" );
                return ret;
            }
            fh_node& NounSynSet()
            {
                static fh_node ret = Node::CreateURI( RDFSCHEMA + "NounSynSet" );
                return ret;
            }
            fh_node& Type()
            {
                static fh_node ret = Node::CreateURI( "http://www.w3.org/1999/02/22-rdf-syntax-ns#type" );
                return ret;
            }
            fh_node& hasGloss()
            {
                static fh_node ret = Node::CreateURI( RDFSCHEMA + "hasGloss" );
                return ret;
            }
            fh_node& hyponymOf()
            {
                static fh_node ret = Node::CreateURI( RDFSCHEMA + "hyponymOf" );
                return ret;
            }
            fh_node& partMeronymOf()
            {
                static fh_node ret = Node::CreateURI( RDFSCHEMA + "partMeronymOf" );
                return ret;
            }
            
            nodelist_t& getSynSet( fh_model m, nodelist_t& ret, const std::string& word )
            {
                fh_node wordNode = Node::CreateURI( WNPFX + "words#" + word );
                
                ObjectIterator end   = ObjectIterator();
                ObjectIterator oiter = m->findSubjects( hasWord(), wordNode );
                for( ; oiter != end; ++oiter )
                {
                    fh_node subject = *oiter;
                    ObjectIterator oiter = m->findObjects( subject, hasSynSet() );
                    for( ; oiter != end; ++oiter )
                    {
                        fh_node synset = *oiter;
                        ret.push_back( synset );
                    }
                }
                return ret;
            }
            
            std::string getNodeID( fh_node n )
            {
                stringstream ret;
                string s = n->getURI()->toString();
                ret << " id=\"" << s.substr( s.rfind("#")+1 ) << "\" ";
                return ret.str();
            }
                
            std::string getWordNodeAsString( fh_node n )
            {
                stringstream ret;
                string s = n->getURI()->toString();
                string sub = s.substr( s.rfind("#")+1 );
                if( !sub.empty() && sub[0] >= '0' && sub[0] <= '9' && sub.length() > WORD_NODE_TEXT_OFFSET )
                    sub = sub.substr( WORD_NODE_TEXT_OFFSET );
    
                ret << sub;
                return ret.str();
            }
                
        };
        
        
    };

    
    /********************************************************************************/
    /********************************************************************************/
    /********************************************************************************/

        struct tryToGetUUIDNode_StringARM
        {
            static void set_tryToGetImplicitTreeSmushHasFailed_forURL( fh_context c, bool v )
                {
                    c->m_tryToGetImplicitTreeSmushHasFailed_forURL = v;
                }
            static bool get_tryToGetImplicitTreeSmushHasFailed_forURL( fh_context c )
                {
                    return c->m_tryToGetImplicitTreeSmushHasFailed_forURL;
                }
            
        };
    
    namespace redlandea
    {
        using namespace Ferris::RDF;
        
        static const std::string ATTR_PREFIX =
        "http://witme.sf.net/libferris.web/rdf/ferris-attr/";
        static const std::string SCHEMA_PREFIX =
        "http://witme.sf.net/libferris.web/rdf/ferris-schema/";
        
        static std::string getURI( const std::string& baseuri,
                                   const std::string& ferrisEAName,
                                   const std::string& pfx = "" )
        {
            std::string ret = baseuri;
            if( !pfx.empty() && starts_with( ferrisEAName, pfx ) )
                ret += ferrisEAName.substr( pfx.length() );
            else
                ret += ferrisEAName;

            return ret;
        }

        std::string getAttrPrefix()
        {
            return ATTR_PREFIX;
        }
        
        
        std::string getPredicateURI( const std::string& rdn )
        {
            if( starts_with( rdn, "schema:" ))
            {
                return getURI( SCHEMA_PREFIX, rdn, "schema:" );
            }
            return getURI( ATTR_PREFIX, rdn );
        }

        std::string stripPredicateURIPrefix( const std::string& pred )
        {
            string ret = pred;
            
            PrefixTrimmer trimmer;
            trimmer.push_back( ATTR_PREFIX );
            trimmer.push_back( SCHEMA_PREFIX );
            ret = trimmer( ret );

            return ret;
        }

        void myrdfSmush( fh_context existingc,
                         fh_context oldc,
                         bool unifyUUIDNodes )
        {
            return myrdfSmush( existingc, oldc->getURL(), unifyUUIDNodes );
        }
        
        void myrdfSmush( fh_context existingc,
                         const std::string& oldpath_maybe_relative,
                         bool unifyUUIDNodes )
        {
            string oldearl = oldpath_maybe_relative;
            if( oldearl.empty() )
            {
                stringstream ss;
                ss << "No old URL given for existingc:" << existingc->getURL() << endl;
                Throw_NoSuchObject( tostr(ss), 0 );
            }
            if( string::npos == oldearl.find("/") )
            {
                string existing_path = existingc->getURL();
                int epos = existing_path.rfind("/");
                existing_path = existing_path.substr( 0, epos );
                existing_path += "/" + oldearl;
                oldearl = existing_path;
            }
            else if( oldearl[0] == '.'  )
            {
                stringlist_t el = Util::parseSeperatedList( existingc->getURL(), '/' );
                stringlist_t ol = Util::parseSeperatedList( oldearl, '/' );
                for( stringlist_t::const_iterator oiter = ol.begin();
                     oiter != ol.end(); ++oiter )
                {
                    if( *oiter == ".." )
                    {
                        stringlist_t::iterator tmp = el.end();
                        --tmp;
                        el.erase( tmp );
                    }
                    else if( *oiter == "." )
                    {
                    }
                    else
                    {
                        el.push_back( *oiter );
                    }
                }

                stringstream ss;
                for( stringlist_t::const_iterator eiter = el.begin();
                     eiter != el.end(); ++eiter )
                {
                    ss << *eiter;
                }

                oldearl = tostr(ss);
            }

            LG_RDF_D << "smushing oldearl:" << oldearl << " existingc:" << existingc->getURL() << endl;
            
            fh_model m = getDefaultFerrisModel();
            fh_node oldURI = Node::CreateURI( oldearl );
            fh_node newURI = Node::CreateURI( existingc->getURL() );

            if( unifyUUIDNodes )
            {
                if( fh_node old_uuidnode = tryToGetUUIDNode( existingc ) )
                {
                    m->erase( Node::CreateURI( existingc->getURL() ),
                              uuidPredNode(),
                              old_uuidnode );
                }
                
                fh_statement partial_statement = new Statement();
                partial_statement->setSubject( oldURI );
                StatementIterator si = m->findStatements( partial_statement );
                for( StatementIterator si_end; si != si_end; ++si )
                {
                    fh_node pred = (*si)->getPredicate();
                    fh_node obj  = (*si)->getObject();
                    
                    m->insert( newURI, pred, obj );
                }
            }
            else
            {
                fh_node n = m->getObject( oldURI, uuidPredNode() );
                if( n )
                {
                    fh_node uuidnode = ensureUUIDNode( existingc );

                    LG_RDF_D << "myrdfSmush()" << endl;
                    LG_RDF_D << "old uuidnode:" << n->toString() << endl;
                    LG_RDF_D << "new uuidnode:" << uuidnode->toString() << endl;
                    
                    bool clone = false;
                    TransitiveRDFNodeCopy trdf_copy( m,m, n, clone );
                    trdf_copy.setNewRootSubject( uuidnode );
                    trdf_copy();
                }
            }
            


            
//             stringlist_t rdf_ea_sl = Util::parseCommaSeperatedList(
//                 getStrAttr( oldchild, "rdf-ea-names", "" ) );
//             stringmap_t rdf_ea_map;
//             if( !rdf_ea_sl.empty() )
//             {
//                 stringlist_t::const_iterator end = rdf_ea_sl.end();
//                 for( stringlist_t::const_iterator si = rdf_ea_sl.begin(); si!=end; ++si )
//                 {
//                     rdf_ea_map[ *si ] = getStrAttr( oldchild, *si, "", true, true );
//                 }
//             }

//             if( !rdf_ea_map.empty() )
//             {
//                 stringmap_t::const_iterator end = rdf_ea_map.end();
//                 for( stringmap_t::const_iterator mi = rdf_ea_map.begin(); mi!=end; ++mi )
//                 {
//                     setStrAttr( child, mi->first, mi->second, true, true );
//                 }
//             }
            
        }
        
        void myrdfSmushUnion( const stringlist_t& srcs )
        {
            fh_model m = getDefaultFerrisModel();

            typedef list< fh_node > nodes_t;
            nodes_t nodes;
            
            for( stringlist_t::const_iterator si = srcs.begin();
                 si != srcs.end(); ++si )
            {
                nodes.push_back( Node::CreateURI( *si ) );
            }

            for( nodes_t::iterator ni = nodes.begin();
                 ni != nodes.end(); ++ni )
            {
                fh_node n = *ni;

                for( nodes_t::iterator nj = nodes.begin();
                     nj != nodes.end(); ++nj )
                {
                    fh_node n2 = *nj;
                    if( n2 == n )
                        continue;

                    bool clone = false;
                    TransitiveRDFNodeCopy trdf_copy( m,m, n2, clone );
                    trdf_copy.setNewRootSubject( n );
                    trdf_copy();
                    

//                     fh_statement partial_statement = new Statement();
//                     partial_statement->setSubject( n2 );
//                     StatementIterator si = m->findStatements( partial_statement );
//                     for( StatementIterator si_end; si != si_end; ++si )
//                     {
//                         fh_node pred = (*si)->getPredicate();
//                         fh_node obj  = (*si)->getObject();

//                         if( !m->getObject( n, pred ) )
//                             m->insert( n, pred, obj );
//                     }
                    
                }
            }
        }

// COMMENT: redlandea should use getPredicateURI() to prepend the URI part.
//         FERRISEXP_API const std::string& getRDFBaseURI();
//         const std::string& getRDFBaseURI()
//         {
//             return RDF::RDF_FERRIS_BASE;
//         }

        string regexEscape( const std::string& s )
        {
            stringstream iss;
            stringstream oss;
            iss << s;
            char ch;
            while( iss >> noskipws >> ch )
            {
                if( ( ch >= 'a' && ch <= 'z' ) 
                    || ( ch >= 'A' && ch <= 'Z' )
                    || ( ch >= '0' && ch <= '9' ) )
                {
                    oss << ch;
                }
                else
                {
                    oss << '\\' << ch;
                }
            }
            return oss.str();
        }


        fh_node getFastSmushByNameNode( fh_SmushSet smushSet, fh_context c )
        {
            stringstream ss;
            ss << smushSet->getName() << "-";
            ss << c->getDirName();
            fh_node subject = Node::CreateURI( getPredicateURI( tostr(ss) ) );
            return subject;
        }
        
        static fh_node createUUIDNode( fh_model m, fh_context c )
        {
            string  earl     = c->getURL();
            fh_node earlnode = Node::CreateURI( earl );
            string     uuid = Util::makeUUID();
            fh_node uuidnode = Node::CreateURI( getPredicateURI( uuid ) );
            m->insert( earlnode, uuidPredNode(), uuidnode );
            
            {
                fh_context p = c->getParent();
                
                if( fh_SmushSet smushSet = p->tryToGetImplicitTreeSmush() )
                {
//                         fh_rex rex = smushSet->getRegex();
//                         if( rex && regex_search( earl, rex, boost::match_any ) )
                    {
                        fh_node subject = getFastSmushByNameNode( smushSet, c );
                        m->insert( subject, smushCacheUUIDPredNode(), uuidnode );
                        LG_RDF_D << "Adding rdf smush cache node." << endl
                                 << " subject:" << subject->toString()
                                 << " pred:" << smushCacheUUIDPredNode()->toString()
                                 << " uuid:" << uuidnode->toString()
                                 << endl;
                    }
                }
            }
            
            fh_node ret = uuidnode;
//            fh_node ret = m->getObject( earlnode, uuidPredNode() );
            setUUIDNodeModificationTime( ret );
            return ret;
        }
        
        fh_node tryToGetUUIDNode( fh_context c )
        {
            fh_model m = getDefaultFerrisModel();

            fh_node earlnode = Node::CreateURI( c->getURL() );
            fh_node uuidnode = m->getObject( earlnode, uuidPredNode() );

            // 
            // If the user has implicit tree smushes setup then we should see if there
            // is a uuidnode attached with the same filename in one of the other paths
            // in a smushset. If we find such a node then we automatically associate
            // it with the current file.
            //
            if( !uuidnode && c->isParentBound() )
            {
                fh_context p = c->getParent();

                LG_RDF_D << "tryToGetUUIDNode() for c:" << c->getURL() << endl;
                 
                if( !tryToGetUUIDNode_StringARM::get_tryToGetImplicitTreeSmushHasFailed_forURL( c ) )
                {
                    tryToGetUUIDNode_StringARM::set_tryToGetImplicitTreeSmushHasFailed_forURL( c, true );

                    if( fh_SmushSet smushSet = p->tryToGetImplicitTreeSmush() )
                    {
                    //
                    // A smush set group leader concept. If the regex is matched then we don't
                    // do any smushing for this URL. This is handy for example when there is
                    // an ingress filesystem where new files are always created (like copied
                    // from a digital camera to there) and there is never any need to smush.
                    //
                    static const string smushLeaderRegexString = getEDBString( FDB_GENERAL,
                                                                               CFG_RDF_GLOBAL_SMUSH_GROUP_LEADER_K,
                                                                               CFG_RDF_GLOBAL_SMUSH_GROUP_LEADER_DEFAULT );
                    if( !smushLeaderRegexString.empty() )
                    {
                        static const fh_rex rex = toregexh( smushLeaderRegexString );
                        if( regex_search( p->getURL(), rex, boost::match_any ) )
                        {
                            LG_RDF_D << "tryToGetUUIDNode() group leader regex matches. Not doing any smushing."
                                     << " smushLeaderRegexString:" << smushLeaderRegexString
                                     << " for c:" << c->getURL() << endl;
                            if( uuidnode )
                                return uuidnode;
                            return createUUIDNode( m, c );
                        }
                    }
                        
                        //
                        // try to hit the new cache directly
                        //
                        {
                            fh_node subject = getFastSmushByNameNode( smushSet, c );
                            LG_RDF_D << "Checking rdf smush cache node." << endl
                                     << " subject:" << subject->toString()
                                     << " pred:" << smushCacheUUIDPredNode()->toString()
                                     << endl;
                            fh_node uuidnode = m->getObject( subject, smushCacheUUIDPredNode() );
                            if( uuidnode )
                            {
                                LG_RDF_D << "FAST Linking uuid:" << uuidnode->toString()
                                         << " to c:" << c->getURL() << endl;
                                m->insert( earlnode, uuidPredNode(), uuidnode );
                                setUUIDNodeModificationTime( uuidnode );
                                return uuidnode;
                            }
                        }
                        
                        
                        
//                        cerr << "tryToGetUUIDNode() implicit smushing for c:" << c->getURL() << endl;
                        
                        stringstream queryiss;
                        queryiss << "PREFIX ferris: <http://witme.sf.net/libferris.web/rdf/ferris-attr/>\n"
                                 << endl
                                 << "SELECT ?uuid ?earl" << endl
                                 << "  WHERE {  " << endl
                                 << "     ?earl ferris:uuid ?uuid . " << endl
                                 << "     FILTER ( " << endl;

                        string regex_string = smushSet->getRegexString();
                        string regexq_filename = Util::replace_all( regexEscape( c->getDirName() ), '\\', "\\\\" );
                        queryiss << "       regex( str(?earl), "
                                 << " \"" << regex_string << "(.*/)*" << regexq_filename << "$\") " << endl;
//                         stringlist_t sl;
//                         const SmushSet::m_smushes_t ss = smushSet->getSmushes();
//                         for( rs<SmushSet::m_smushes_t> r( ss ); r; ++r )
//                         {
//                             sl.push_back( si->first );
//                         }
//                         std::string compregex = MakeCompositeRegexString( sl );

                        
//                         bool virgin = true;
//                         const SmushSet::m_smushes_t ss = smushSet->getSmushes();
//                         SmushSet::m_smushes_t::const_iterator si = ss.begin();
//                         SmushSet::m_smushes_t::const_iterator se = ss.end();
//                         for( ; si != se ; ++si )
//                         {
//                             const string& regex_string = si->first;
//                             fh_regex r = si->second;
//                             string regexq_filename = Util::replace_all( regexEscape( c->getDirName() ), '\\', "\\\\" );

//                             if( !virgin )
//                                 queryiss << " || " ;
//                             virgin = false;
                        

//                             queryiss << "       regex( str(?earl), "
//                                      << " \"" << regex_string << "(.*/)*" << regexq_filename << "$\") " << endl;
//                         }

                        queryiss << " ) } LIMIT 1  " << endl;
                        
                        string query = queryiss.str();
                        LG_RDF_D << "Implicit Tree Smush query:" << endl << query << endl;

                        //
                        //
                        //
                        LG_RDF_W << "SPARQL is too slow to do at runtime. see apps/rdf/ferris-myrdf-inference" << endl;
//                         BindingsIterator iter = m->findBindings( query );
//                         BindingsIterator e;
//                         for( ; iter != e ; ++iter )
//                         {
//                             const BindingsIterator::binding& b = *iter;
//                             uuidnode = b.getByName( "uuid" );
//                             //
//                             // attach the found uuidnode with the URL.
//                             //
//                             LG_RDF_D << "Linking uuid:" << uuidnode->toString()
//                                      << " to c:" << c->getURL() << endl;
//                             m->insert( earlnode, uuidPredNode(), uuidnode );
//                             setUUIDNodeModificationTime( uuidnode );
//                             return uuidnode;
//                         }

                        // A once off hit making sure that all directories in a tree smush
                        // will have a uuid node.
                        LG_RDF_D << "tryToGetUUIDNode() calling ensure for c:" << c->getURL() << endl;
                        return ensureUUIDNode( c );
                    }
                }
            }

            return uuidnode;
        }

        fh_node ensureUUIDNode( fh_model m, fh_context c )
        {
            fh_node uuidnode = tryToGetUUIDNode( c );
            if( uuidnode )
                return uuidnode;

            return createUUIDNode( m, c );
        }


        fh_node ensureUUIDNode( fh_context c )
        {
            fh_model m = getDefaultFerrisModel();
            return ensureUUIDNode( m, c );
        }

        
        fh_node& uuidMTimeNode()
        {
            static fh_node ret = 0;
            if( !ret )
                ret = Node::CreateURI( getPredicateURI( "mtime" ) );
            return ret;
        }
        
        time_t getUUIDNodeModificationTime( ::Ferris::RDF::fh_node n )
        {
            fh_model m = getDefaultFerrisModel();
            fh_node pred = uuidMTimeNode();
            fh_node obj = m->getObject( n, pred );
            if( obj )
            {
                time_t ret = toType<time_t>(obj->toString());
                LG_RDF_D << "getUUIDNodeModificationTime() ret:" << ret
                         << " for:" << n->toString() << endl;
                return ret;
            }
            else
            {
                LG_RDF_D << "getUUIDNodeModificationTime() no data for:" << n->toString() << endl;
            }
            return 0;
        }
        
        void setUUIDNodeModificationTime( ::Ferris::RDF::fh_node n, time_t t )
        {
            fh_model m = getDefaultFerrisModel();
            if( !t )
            {
                t = Time::getTime();
            }

            LG_RDF_D << "setUUIDNodeModificationTime() n:" << n->toString()
                     << " t:" << t << endl;
//             cerr << "setUUIDNodeModificationTime() n:" << n->toString()
//                  << " t:" << t << endl;
//             BackTrace();
            
            fh_node pred = uuidMTimeNode();
            fh_node obj = Node::CreateLiteral( tostr(t) );
            m->set( n, pred, obj );
            m->sync();
        }

        void setRedlandEASubjectNode( ::Ferris::RDF::fh_node n, time_t t )
        {
            fh_model m = getDefaultFerrisModel();

            fh_node uuidnode = m->getSubject( uuidOutOfBandPredNode(), n );
            if( uuidnode )
                setUUIDNodeModificationTime( uuidnode, t );
        }
        
        
        
        
        fh_node tryToGetRedlandEASubjectNode( fh_context c )
        {
            fh_model m = getDefaultFerrisModel();
            fh_node ret = tryToGetUUIDNode( c );
            if( ret )
            {
                ret = m->getObject( ret, uuidOutOfBandPredNode() );
            }
            return ret;
        }
        
        fh_node ensureRedlandEASubjectNode( fh_context c )
        {
            fh_node ret = tryToGetRedlandEASubjectNode( c );
            if( ret )
                return ret;
            
            fh_model m = getDefaultFerrisModel();
            fh_node uuidnode = ensureUUIDNode( c );
//            fh_node mdnode = Node::CreateURI( getPredicateURI( Util::makeUUID() ) );
            fh_node mdnode = Node::CreateBlankNode();
            m->set( uuidnode, uuidOutOfBandPredNode(), mdnode );
            setUUIDNodeModificationTime( uuidnode );
            return mdnode;
        }
        
        

        fh_node& uuidPredNode()
        {
            static fh_node ret = 0;
            if( !ret )
                ret = Node::CreateURI( getPredicateURI( "uuid" ) );
            return ret;
        }

        fh_node& smushCacheUUIDPredNode()
        {
            static fh_node ret = 0;
            if( !ret )
                ret = Node::CreateURI( getPredicateURI( "smushcacheuuid" ) );
            return ret;
        }

        fh_node& uuidOutOfBandPredNode()
        {
            static fh_node ret = 0;
            if( !ret )
                ret = Node::CreateURI( getPredicateURI( "out-of-band-ea" ) );
            return ret;
        }
        
        /********************************************************************************/
        /********************************************************************************/
        /********************************************************************************/

        SmushSet::~SmushSet()
        {
        }
            

        const std::string&
        SmushSet::getName() const
        {
            return m_name;
        }
        
        void
        SmushSet::setName( const std::string& n )
        {
            m_treeSmushing->updateSmushName( m_name, n );
            m_name = n;
        }
        

        const SmushSet::m_smushes_t&
        SmushSet::getSmushes() const
        {
            return m_smushes;
        }

        const string&
        SmushSet::getRegexString()
        {
            return m_regexString;
        }

        fh_rex
        SmushSet::getRegex()
        {
            if( !m_rex )
            {
                if( !getRegexString().empty() )
                {
                    m_rex = toregexh( getRegexString() );
                }
            }
            return m_rex;
        }
        
        
        void
        SmushSet::setSmushes( m_smushes_t& v )
        {
            m_smushes = v;
            stringlist_t sl;
            for( rs<m_smushes_t> r(m_smushes); r; ++r )
                sl.push_back( r->first );
            m_regexString = MakeCompositeRegexString( sl );
        }
        
        SmushSet::SmushSet( fh_TreeSmushing p, const std::string& name )
            :
            m_treeSmushing( p ),
            m_name( name ),
            m_rex( 0 )
        {
        }
        
            
        TreeSmushing::TreeSmushing()
            :
            m_shouldSave( false )
        {
        }
        
        TreeSmushing::~TreeSmushing()
        {
        }

        const TreeSmushing::m_smushSets_t&
        TreeSmushing::getAll() const
        {
            return m_smushSets;
        }
        
        
        fh_SmushSet
        TreeSmushing::getSmush( const std::string& n ) const
        {
            m_smushSets_t::const_iterator ci = m_smushSets.find( n );
            if( ci != m_smushSets.end() )
                return ci->second;
            return 0;
        }
        
        void
        TreeSmushing::swap( fh_TreeSmushing other )
        {
            m_smushSets.swap( other->m_smushSets );
        }
        
        fh_SmushSet
        TreeSmushing::newSmush( const std::string& name )
        {
            m_smushSets[ name ] = new SmushSet( this, name );
            return m_smushSets[ name ];
        }

        std::string
        TreeSmushing::getBoostSerializePath()
        {
            return CleanupURL( "~/.ferris/tree-smushing.boost" );
        }

        template< class ArchiveClassT >
        void
        TreeSmushing::sync( ArchiveClassT& archive )
        {
            m_smushSets_t::iterator ci = m_smushSets.begin();
            m_smushSets_t::iterator  e = m_smushSets.end();

            long sz = m_smushSets.size();
            archive << sz;
            for( ; ci != e ; ++ci )
            {
                archive << ci->first;

                fh_SmushSet ss = ci->second;
                long sz = ss->m_smushes.size();
                archive << sz;

                SmushSet::m_smushes_t::iterator si = ss->m_smushes.begin();
                SmushSet::m_smushes_t::iterator se = ss->m_smushes.end();
                for( ; si != se; ++si )
                {
                    archive << si->first;
                    const Regex& rr = *(GetImpl(si->second));
                    archive << rr;
                }
            }
        }
        
        void
        TreeSmushing::sync()
        {
            if( m_shouldSave )
            {
                {
                    string fname = getBoostSerializePath() + ".txt";
                    std::ofstream ofs( fname.c_str() );
                    boost::archive::text_oarchive archive( ofs );
                    sync( archive );
                }
                
                {
                    std::ofstream ofs( getBoostSerializePath().c_str() );
                    boost::archive::binary_oarchive archive( ofs );
                    sync( archive );
                }
            }
        }
        

        TreeSmushing::TreeSmushing( bool shouldSave )
            :
            m_shouldSave( shouldSave )
        {
            if( m_shouldSave )
            {
                if( !access( getBoostSerializePath().c_str(), R_OK ) )
                {
                    std::ifstream ifs( getBoostSerializePath().c_str() );
                    boost::archive::binary_iarchive archive( ifs );
                    
                    long sz = 0;
                    archive >> sz;
                    for( long iter = 0; iter < sz; ++iter )
                    {
                        string name;
                        archive >> name;
                        fh_SmushSet ss = new SmushSet( this, name );
                        long ss_sz = 0;
                        archive >> ss_sz;
                        
                        SmushSet::m_smushes_t smushes;
                        for( long ss_iter = 0; ss_iter < ss_sz; ++ss_iter )
                        {
                             string s;
                             archive >> s;
                             
                             fh_regex r = new Regex("");
                             Regex& rr = *(GetImpl(r));
                             archive >> rr;
                             smushes.insert( make_pair( s, r ) );
                         }
                         ss->setSmushes( smushes );
                        m_smushSets.insert( make_pair( name, ss ) );
                    }
                }
            }
        }
        
        void
        TreeSmushing::updateSmushName( const std::string& oldname, const std::string& newname )
        {
            m_smushSets_t::iterator ci = m_smushSets.find( oldname );
            if( ci != m_smushSets.end() )
            {
                fh_SmushSet d = ci->second;
                m_smushSets.erase( ci );
                m_smushSets[ newname ] = d;
            }
        }
        
        
        fh_TreeSmushing getDefaultImplicitTreeSmushing()
        {
            static fh_TreeSmushing ret = 0;
            if( !ret )
            {
                ret = new TreeSmushing( true );
            }
            return ret;
        }

        /********************************************************************************/
        /********************************************************************************/
        /********************************************************************************/
        
        
    };

};

