/******************************************************************************
*******************************************************************************
*******************************************************************************

    libferris
    Copyright (C) 2001 Ben Martin

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

    For more details see the COPYING file in the root directory of this
    distribution.

    $Id: RootContext_private.hh,v 1.3 2007/01/22 21:30:10 ben Exp $

*******************************************************************************
*******************************************************************************
******************************************************************************/

#ifndef _ALREADY_INCLUDED_FERRIS_ROOT_CONTEXT_PRIV_H_
#define _ALREADY_INCLUDED_FERRIS_ROOT_CONTEXT_PRIV_H_

#include <Ferris/HiddenSymbolSupport.hh>
#include <Ferris/Ferris.hh>
#include <Ferris/Context.hh>

namespace Ferris
{
    

    class FERRISEXP_DLLLOCAL RootContext : public ChainedViewContext
    {
        typedef ChainedViewContext _Base;
        typedef Context _NonChainedBase;  //< lowest superclass which is not under ChainedViewContext
        
        friend class RootContextFactory;

        Context* priv_CreateContext( Context* parent, std::string rdn );
    
    protected:

        std::set< std::string >& getPreferLocalAttributeNames();
        bool isAttributeLocal( const std::string& s );
        virtual std::string private_getStrAttr( const std::string& rdn,
                                                const std::string& def = "",
                                                bool getAllLines = false ,
                                                bool throwEx = false );
        
    public:
        RootContext( Context* parent,
                     Context* delegate,
                     const std::string& rdn = "" );

        
        virtual void read( bool force = 0 );
        virtual const std::string& getDirName() const;
        virtual std::string getDirPath() throw (FerrisParentNotSetError);
        virtual std::string getURL();
        virtual std::string getURLScheme();

        virtual fh_attribute getAttribute( const std::string& rdn ) throw( NoSuchAttribute );
        virtual AttributeNames_t& getAttributeNames( AttributeNames_t& ret );
        virtual int  getAttributeCount();
        virtual bool isAttributeBound( const std::string& rdn,
                                       bool createIfNotThere = true
            ) throw( NoSuchAttribute );
        
//        virtual void OnExists ( NamingEvent_Exists* ev,  std::string olddn, std::string newdn );
    };

};
#endif
