/******************************************************************************
*******************************************************************************
*******************************************************************************

    libferris
    Copyright (C) 2001 Ben Martin

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

    For more details see the COPYING file in the root directory of this
    distribution.

    $Id: SchemaManips.cpp,v 1.1 2005/07/04 08:56:43 ben Exp $

*******************************************************************************
*******************************************************************************
******************************************************************************/

#include <SchemaManips.hh>

namespace Ferris
{
    justification::justification( fh_context schemac )
        : one_arg_manip< fh_context >( justification::fct, schemac )
    {}
        
    void
    justification::fct( std::ios_base& ib, fh_context schemac )
    {
        std::string js = schemac
            ? getStrAttr( schemac, SCHEMA_JUSTIFICATION, "left" )
            : "left";
        
        if( js == "right" )
            ib.setf( std::ios_base::right, std::ios_base::adjustfield );
        else
            ib.setf( std::ios_base::left, std::ios_base::adjustfield );
    }

    justificationC::justificationC( fh_context& c, const std::string& eaname )
        : two_arg_manip< fh_context, std::string >( justificationC::fct, c, eaname )
    {}
    void
    justificationC::fct( std::ios_base& ib,
                         const fh_context& c,
                         const std::string& eaname )
    {
        std::string js = getSchemaJustification( c, eaname, "left" );
        
        if( js == "right" )
            ib.setf( std::ios_base::right, std::ios_base::adjustfield );
        else
            ib.setf( std::ios_base::left, std::ios_base::adjustfield );
    }
    
    width::width( fh_context schemac )
        : one_arg_manip< fh_context >( width::fct, schemac )
    {}
    void
    width::fct( std::ios_base& ib, fh_context schemac )
    {
        int wid = schemac
            ? toint(getStrAttr( schemac, SCHEMA_DEFAULT_DISPLAY_WIDTH, "0" ))
            : 0;
        
        if( wid )
        {
            ib.width( wid );
        }
    }
    widthC::widthC( fh_context c, const std::string& eaname )
        : two_arg_manip< fh_context, std::string >( widthC::fct, c, eaname )
    {}
    void
    widthC::fct( std::ios_base& ib,
                 const fh_context& c,
                 const std::string& eaname )

    {
        int wid = getSchemaDisplayWidth( c, eaname, 0 );
        
        if( wid )
        {
            ib.width( wid );
        }
    }
};
