/******************************************************************************
*******************************************************************************
*******************************************************************************

    libferris
    Copyright (C) 2001 Ben Martin

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

    For more details see the COPYING file in the root directory of this
    distribution.

    $Id: EditStringList.hh,v 1.1 2005/07/04 08:55:52 ben Exp $

*******************************************************************************
*******************************************************************************
******************************************************************************/

#ifndef _ALREADY_INCLUDED_FERRISUI_EDIT_STRING_LIST_H_
#define _ALREADY_INCLUDED_FERRISUI_EDIT_STRING_LIST_H_

#include <Ferris/HiddenSymbolSupport.hh>

#include <gtk/gtk.h>

#include <Ferris/TypeDecl.hh>

#include <string>

namespace FerrisUI
{
    class FERRISEXP_API EditStringList
        :
        public Ferris::Handlable
    {
        friend void EditStringList_add_cb( GtkButton *button, gpointer user_data );
        friend void EditStringList_del_cb( GtkButton *button, gpointer user_data );
        friend void EditStringList_clear_cb( GtkButton *button, gpointer user_data );
        friend void EditStringList_edited_cb (GtkCellRendererText *cell, gchar *path_string,
                                              gchar *new_text, gpointer udata);

    protected:

        enum {
            C_STRING_COLUMN=0,
            C_COLUMN_COUNT
        };
        GtkTreeStore*      w_treemodel;
        GtkWidget*         w_treeview;
        GtkTreeViewColumn* w_cols[C_COLUMN_COUNT];
        GtkWidget*         w_baseWidget;
        
        std::string m_columnLabel;
        std::string m_descriptionLabel;

    public:

        EditStringList();
        virtual ~EditStringList();

        /**
         * Set the description string for this string list
         * call before getWidget()
         */
        void setDescriptionLabel( const std::string& s );

        /**
         * set the name for the column header
         * call before getWidget()
         */
        void setColumnLabel( const std::string& s );

        /**
         * Get the current contents of the string list
         * call AFTER getWidget()
         */
        Ferris::stringlist_t getStringList();

        /**
         * set the content of the string list
         * call AFTER getWidget()
         */
        void setStringList( const Ferris::stringlist_t& sl );

        /**
         * Get the GTK+ view for this string editing component.
         */
        GtkWidget* getWidget();

        /**
         * clear the stringlist
         * call AFTER getWidget()
         */
        void clear();

        /**
         * add a new blank item
         * call AFTER getWidget()
         */
        GtkTreeIter appendNewBlankItem();

    };

    FERRIS_SMARTPTR( EditStringList, fh_editstringlist );
    
};
#endif
