/******************************************************************************
*******************************************************************************
*******************************************************************************

    fcat
    Copyright (C) 2002 Ben Martin

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

    For more details see the COPYING file in the root directory of this
    distribution.

    $Id: fcat.cpp,v 1.5 2007/03/14 21:30:12 ben Exp $

*******************************************************************************
*******************************************************************************
******************************************************************************/

/*
 * return 0 for success
 * return 1 for generic error
 * return 5 for not modified
 * 
 *
 *
 *
 *
bash -c '/ferris/apps/cat/fcat Makefile 2>/dev/null | wc -l'



*/


#include <Ferris.hh>
#include <popt.h>
#include <unistd.h>

using namespace std;
using namespace Ferris;

const string PROGRAM_NAME = "fcat";

void usage(poptContext optCon, int exitcode, char *error, char *addl)
{
    poptPrintUsage(optCon, stderr, 0);
    if (error) fprintf(stderr, "%s: %s0", error, addl);
    exit(exitcode);
}

unsigned long OutputHeadersOnFD = 0;

/********************************************************************************/
/********************************************************************************/
/********************************************************************************/

void OutputHeadersOnFD_cb( fh_context c, const stringset_t& strset )
{
    try
    {
        cerr << "OutputHeadersOnFD_cb()" << endl;
        fh_stringstream ss;
        ss << "<?xml version=\"1.0\" encoding=\"UTF-8\" standalone=\"no\"?> " << endl;
        ss << "<headerinfo> " << endl;
        for( stringset_t::const_iterator iter = strset.begin(); iter != strset.end(); ++iter )
        {
            string k = *iter;
            cerr << "OutputHeadersOnFD_cb() k:" << k <<  endl;
            string v = getStrAttr( c, k, "" );
            ss << "<" << k << ">";
            ss << v;
            ss << "</" << k << ">" << endl;
        }
        ss << "</headerinfo> " << endl;

        fh_ostream oss = Factory::MakeFdOStream( OutputHeadersOnFD );
        oss << tostr(ss) << flush;
    }
    catch( exception& e )
    {
        cerr << "OutputHeadersOnFD_cb() e:" << e.what() << endl;
    }
}

/********************************************************************************/
/********************************************************************************/
/********************************************************************************/

int main( int argc, char** argv )
{
    int exit_status = 0;
    
    try
    {
        const char* SourceAttrName         = "content";
        const char* DownloadIfMtimeSinceStr= 0;
        unsigned long DownloadIfMtimeSince = 0;
        unsigned long Verbose              = 0;

        struct poptOption optionsTable[] =
            {
                { "verbose", 'v', POPT_ARG_NONE, &Verbose, 0,
                  "show what is happening", "" },

                { "download-if-mtime-since", 0, POPT_ARG_INT, &DownloadIfMtimeSince, 0,
                  "only transfer file content if it has been modified after given long time", "" },

                { "output-headers-on-fd", 0, POPT_ARG_INT, &OutputHeadersOnFD, 0,
                  "when header info is received output an XML docuemnt on the given already open fd.",
                  "" },
                
                { "download-if-mtime-since-string", 0, POPT_ARG_STRING,
                  &DownloadIfMtimeSinceStr, 0,
                  "only transfer file content if it has been modified after given string time", "" },

                { "src-attr", 'a', POPT_ARG_STRING,
                  &SourceAttrName, 0,
                  "cat an EA rather than the content itself", "" },
                
//             { "target-directory", 0, POPT_ARG_STRING, &DstNameCSTR, 0,
//               "Specify destination explicity, all remaining URLs are assumed to be source files",
//               "DIR" },

                FERRIS_POPT_OPTIONS
                POPT_AUTOHELP
                POPT_TABLEEND
            };
        poptContext optCon;

        optCon = poptGetContext(PROGRAM_NAME.c_str(), argc, (const char**)argv, optionsTable, 0);
        poptSetOtherOptionHelp(optCon, "[OPTIONS]* src1 src2 ...");

        /* Now do options processing */
        char c=-1;
        while ((c = poptGetNextOpt(optCon)) >= 0)
        {
//         switch (c) {
//         }
        }

        if (argc < 2)
        {
            poptPrintHelp(optCon, stderr, 0);
            exit(1);
        }


        stringlist_t srcs;
        srcs = expandShellGlobs( srcs, optCon );

        stringlist_t::iterator srcsiter = srcs.begin();
        stringlist_t::iterator srcsend  = srcs.end();
        for( ; srcsiter != srcsend ; ++srcsiter )
        {
            string srcURL = *srcsiter;
            fh_istream iss;
            fh_ostream oss = Factory::MakeFdOStream( STDOUT_FILENO );
            
            try
            {
//                cerr << "srcURL :" << srcURL << endl;
                
                if( srcURL == "-" )
                {
                    iss = Factory::MakeFdIStream( STDIN_FILENO );
                }
                else
                {
                    fh_context c = Resolve( srcURL );
                    if( DownloadIfMtimeSince )
                        setStrAttr( c, "download-if-mtime-since",
                                    tostr(DownloadIfMtimeSince),
                                    false, false );
                    if( DownloadIfMtimeSinceStr )
                    {
                        try
                        {
                            cerr << "DownloadIfMtimeSinceStr:" << DownloadIfMtimeSinceStr << endl;
                            cerr << "c:" << c->getURL() << endl;
                            setStrAttr( c, "download-if-mtime-since-display",
                                        DownloadIfMtimeSinceStr,
                                        false, true );
                        }
                        catch( exception& e )
                        {
                            cerr << "DownloadIfMtimeSinceStr e:" << e.what() << endl;
                        }
                    }

                    SigC::Connection hdrcon;
                    
                    if( OutputHeadersOnFD )
                    {
                        hdrcon = c->getContextEvent_Headers_Received_Sig().connect(
                            SigC::slot( OutputHeadersOnFD_cb ) );
                    }

                    /*
                     * Either cat an EA or the file itself,
                     * SourceAttrName==content by default
                     */
                    fh_attribute a = c->getAttribute( SourceAttrName );
                    iss = a->getIStream();
                    
                    if( OutputHeadersOnFD )
                    {
                        hdrcon.disconnect();
                    }
                    
                }

//                 iss.exceptions( ios_base::badbit | ios_base::failbit );
//                 cerr << "set exceptions to on...." << endl;
                
                copy( istreambuf_iterator<char>(iss), istreambuf_iterator<char>(),
                      ostreambuf_iterator<char>(oss));
                oss << flush;

//                 cerr << " iss is good():"<< iss.good() << endl;
//                 cerr << " iss is eof():"<< iss.eof() << endl;
//                 cerr << " iss is state:"<< iss.rdstate() << endl;
//                 if ( iss.rdstate() & ifstream::failbit )
//                     cerr << " iss has fail bit set." << endl;
//                 if ( iss.rdstate() & ifstream::badbit )
//                     cerr << " iss has bad bit set." << endl;

                
                if( haveIOError( iss ) )
                {
                    cerr << getIOErrorDescription( iss, srcURL ) << endl;
                }
                if( haveIOError( oss ) )
                {
                    cerr << getIOErrorDescription( oss, "<STDOUT>" ) << endl;
                }
            }
            catch( ContentNotModified& e )
            {
                cerr << "not modified:" << srcURL << endl;
                return 5;
            }
            catch( exception& e )
            {
                cerr << "error:" << e.what() << endl;
                exit_status = 1;
            }
        }
    }
    catch( exception& e )
    {
        cerr << "error:" << e.what() << endl;
        exit(1);
    }
    return exit_status;
}


