/******************************************************************************
*******************************************************************************
*******************************************************************************

    ferris cp
    Copyright (C) 2001 Ben Martin

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

    For more details see the COPYING file in the root directory of this
    distribution.

    $Id: ferriscp.cpp,v 1.4 2007/12/11 21:30:13 ben Exp $

*******************************************************************************
*******************************************************************************
******************************************************************************/

#include <string>
#include <list>
#include <set>
#include <vector>
#include <algorithm>
#include <iomanip>

#include <FerrisCopy.hh>
#include <FerrisCopy_private.hh>

#include <popt.h>

#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>

using namespace std;
using namespace Ferris;


const string PROGRAM_NAME = "ferriscp";


void usage(poptContext optCon, int exitcode, char *error, char *addl)
{
    poptPrintUsage(optCon, stderr, 0);
    if (error) fprintf(stderr, "%s: %s0", error, addl);
    exit(exitcode);
}

/******************************************************************************/
/******************************************************************************/
/******************************************************************************/
/******************************************************************************/
/******************************************************************************/
/******************************************************************************/
/******************************************************************************/
/*******************************************************************************/
/*******************************************************************************/
/*******************************************************************************/
/*******************************************************************************/


int main( int argc, const char** argv )
{
    fh_cp_tty obj = new FerrisCopy_TTY();

    const char* DstNameCSTR              = 0;
    unsigned long ShowMeter              = 0;

    struct poptOption optionsTable[] =
        {
            { "show-progress-meter", 0, POPT_ARG_NONE, &ShowMeter, 0,
              "Show a one line progress meter", "" },

            { "target-directory", 0, POPT_ARG_STRING, &DstNameCSTR, 0,
              "Specify destination explicity, all remaining URLs are assumed to be source files",
              "DIR" },

            FERRIS_COPY_OPTIONS( obj )
            FERRIS_POPT_OPTIONS
            POPT_AUTOHELP
            POPT_TABLEEND
        };
    poptContext optCon;

    optCon = poptGetContext(PROGRAM_NAME.c_str(), argc, argv, optionsTable, 0);
    poptSetOtherOptionHelp(optCon, "[OPTIONS]* src1 src2 ... dst");

    /* Now do options processing */
    char c=-1;
    while ((c = poptGetNextOpt(optCon)) >= 0)
    {
//         switch (c) {
//         }
    }

    obj->getPoptCollector()->ArgProcessingDone( optCon );

    if (argc < 3)
    {
        poptPrintHelp(optCon, stderr, 0);
        exit(1);
    }

    try
    {
        typedef stringlist_t srcs_t;
        srcs_t srcs;

        srcs = expandShellGlobs( srcs, optCon );
        
        /*
         * Setup the destination from either explicit command line or
         * last arg
         */
        string DstName;
        if( DstNameCSTR )
        {
            obj->setDstIsDirectory( true );
            DstName = DstNameCSTR;
        }
        else
        {
            DstName = srcs.back();
            srcs.pop_back();
        }

        /* TTY version */
        obj->setShowMeter( ShowMeter );

        /* Common version */
//        cerr << "dst:" << DstName << endl;
        obj->setDstURL( DstName );

        /************************************************************/
        /************************************************************/
        /************************************************************/
        
        for( srcs_t::iterator iter = srcs.begin(); iter != srcs.end(); ++iter )
        {
            string SrcName = *iter;
//            cerr << "src:" << SrcName << endl;

            obj->setSrcURL( SrcName );
            obj->copy();
        }
    }
    catch( NoSuchContextClass& e )
    {
        cerr << "Invalid context class given e:" << e.what() << endl;
        exit(1);
    }
    catch( exception& e )
    {
        cerr << "Error: " << e.what() << endl;
        exit(1);
    }
    


    
    poptFreeContext(optCon);
    return 0;
}
