/******************************************************************************
*******************************************************************************
*******************************************************************************

    eaindex add command line client
    Copyright (C) 2003 Ben Martin

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

    For more details see the COPYING file in the root directory of this
    distribution.

    $Id: feaindexadd.cpp,v 1.4 2007/11/26 21:30:14 ben Exp $

*******************************************************************************
*******************************************************************************
******************************************************************************/

/*
 * return 0 for success
 * return 1 for generic error
*/

#include <Ferris/Ferris.hh>
#include <Ferris/Ferrisls.hh>
#include <Ferris/FullTextIndexer.hh>
#include <Ferris/EAIndexer.hh>
#include <Ferris/EAIndexer_private.hh>
#include <popt.h>
#include <unistd.h>

using namespace std;
using namespace Ferris;
using namespace FullTextIndex;

const string PROGRAM_NAME = "feaindexadd";

void usage(poptContext optCon, int exitcode, char *error, char *addl)
{
    poptPrintUsage(optCon, stderr, 0);
    if (error) fprintf(stderr, "%s: %s0", error, addl);
    exit(exitcode);
}

/********************************************************************************/
/********************************************************************************/
/********************************************************************************/


void progressf( fh_context c, streamsize bytesDone, streamsize totalBytes )
{
    cerr << "c:" << c->getURL() << " " << bytesDone << " / " << totalBytes << endl;
}


/********************************************************************************/
/********************************************************************************/
/********************************************************************************/

Time::Benchmark m_totalRuntime( "Total indexing time", false );

class Ferrisls_feaindexadd_display
    :
    public Ferrisls_display
{
    typedef Ferrisls_display _Base;

    EAIndex::fh_idx m_idx;
    EAIndex::fh_docindexer m_indexer;
    bool m_verbose;
    bool m_showTotals;
    int m_contextCount;
    
    virtual void workStarting();
    virtual void workComplete();
    virtual void EnteringContext(fh_context ctx);
    virtual void LeavingContext(fh_context ctx);
        
protected:

    virtual void ShowAttributes( fh_context ctx );
    virtual void PrintEA( fh_context ctx,
                          int i,
                          const std::string& attr,
                          const std::string& EA );
public:

    Ferrisls_feaindexadd_display( EAIndex::fh_idx idx )
        :
        m_idx( idx ),
        m_contextCount( 0 )
        {
            m_indexer = EAIndex::Factory::makeDocumentIndexer( m_idx );
        }

    void sync()
        {
            m_idx->sync();
        }
    
    void printTotals()
        {
            if( !m_showTotals )
                return;
        }
    

    void setVerbose( bool v )
        {
            m_verbose = v;
        }

    void setAddEvenIfAlreadyCurrent( bool v )
        {
            m_indexer->setAddEvenIfAlreadyCurrent( v );
        }
    

    void setShowTotals( bool v )
        {
            m_showTotals = v;
        }
    
    void setIndex( EAIndex::fh_idx idx )
        {
            m_idx = idx;
        }
    
    void setIgnoreEANames( const std::string& s )
        {
            m_indexer->setIgnoreEANames( s );
        }
    
    void appendIgnoreEANames( const std::string& s )
        {
            m_indexer->appendIgnoreEANames( s );
        }
    void appendIgnoreEARegexs( const std::string& s )
        {
            m_indexer->appendIgnoreEARegexs( s );
        }
    
    void setMaximunValueSize( std::streamsize sz )
        {
            m_indexer->setMaximunValueSize( sz );
        }

    void setDontCheckIfAlreadyThere( bool v )
        {
            m_indexer->setDontCheckIfAlreadyThere( v );
        }
    
    void setEANamesToIndex( const stringlist_t& sl )
        {
            m_indexer->clearIgnoreEARegexs();
            m_indexer->clearIgnoreEANames();
            m_indexer->setEANamesToIndex( sl );
        }
    int getFilesIndexedCount()
        {
            return m_indexer->getFilesIndexedCount();
        }
    void setRetireNonExistentDocumentsToMuliversion( bool v )
        {
            m_indexer->setRetireNonExistentDocumentsToMuliversion( v );
        }
    
};

FERRIS_SMARTPTR( Ferrisls_feaindexadd_display, fh_eaindexadd_display );

void
Ferrisls_feaindexadd_display::PrintEA( fh_context ctx,
                               int i,
                               const std::string& attr,
                               const std::string& EA )
{
}

void
Ferrisls_feaindexadd_display::ShowAttributes( fh_context ctx )
{
    if( m_verbose )
        cerr << "Indexing context:" << ctx->getURL() << endl;

    try
    {
        m_indexer->addContextToIndex( ctx );
    }
    catch( exception& e )
    {
        cerr << "feaindexadd error:" << e.what() << endl;
    }
    
    if( m_showTotals )
    {
        cerr << "ShowAttributes() ADDING CONTEXT ctx:" << ctx->getURL() << endl;
        ++m_contextCount;
    }
    
}

void
Ferrisls_feaindexadd_display::workStarting()
{
}

void
Ferrisls_feaindexadd_display::workComplete()
{
}

void
Ferrisls_feaindexadd_display::EnteringContext(fh_context ctx)
{
    if( m_verbose )
        cerr << "Entering context:" << ctx->getURL() << endl;

    m_indexer->EnteringContext( ctx );
}

void
Ferrisls_feaindexadd_display::LeavingContext(fh_context ctx)
{
    if( m_verbose )
        cerr << "Leaving context:" << ctx->getURL() << endl;

    m_indexer->LeavingContext( ctx );
}


/********************************************************************************/
/********************************************************************************/
/********************************************************************************/
/********************************************************************************/
/********************************************************************************/
/********************************************************************************/

int exit_status = 0;
int TotalFilesDoneCount = 0;
unsigned long AddEvenIfAlreadyCurrent = 0;

void addToIndexFromFileList( EAIndex::fh_idx& idx, fh_istream& fiss )
{
    string srcURL;
            
    EAIndex::fh_docindexer indexer = EAIndex::Factory::makeDocumentIndexer( idx );
    indexer->setAddEvenIfAlreadyCurrent( AddEvenIfAlreadyCurrent );
    while( getline( fiss, srcURL ) )
    {
        try
        {
            if( srcURL.empty() )
                continue;

            fh_context c = Resolve( srcURL );

            if( idx->getIndexMethodSupportsIsFileNewerThanIndexedVersion() )
            {
                if( !idx->isFileNewerThanIndexedVersion( c ) )
                {
                    cerr << "Skipping:" << srcURL << endl;
                    continue;
                }
            }
            
            cerr << "ADDING:" << srcURL << endl;
            indexer->addContextToIndex( c );
        }
        catch( exception& e )
        {
            cerr << "----------------------" << endl;
            cerr << "for: " << srcURL << endl;
            cerr << "cought error:" << e.what() << endl;
            exit_status = 1;
        }
        ++TotalFilesDoneCount;
    }
}


int main( int argc, char** argv )
{
    Ferrisls ls;
    
    try
    {
//        const char* CreateTypeName_CSTR    = 0;
        const char* IndexPath_CSTR         = 0;
        const char* FilelistFile_CSTR      = 0;
        unsigned long FilelistStdin        = 0;
        const char* IgnoreEANames          = 0;
        const char* IgnoreEANamesAppend    = 0;
        const char* IgnoreEARegexsAppend   = 0;
        const char* EAToIndex              = 0;
        unsigned long Verbose              = 0;
        unsigned long EAIndexMaxValueSize
            = toint( EAIndex::GET_EAINDEX_MAX_VALUE_SIZE_TO_INDEX_DEFAULT() );
        unsigned long DontCheckIfAlreadyThere = 0;
        unsigned long ListDirectoryNodeOnly = 0;
        unsigned long ForceReadRootDirectoryNodes = 0;
        unsigned long RecursiveList   = 0;
        const char* FilterStringCSTR = 0;
        unsigned long RetireNonExistentDocumentsToMuliversion = 0;

        unsigned long ShowTotals = 0;
        
        struct poptOption optionsTable[] =
            {
                { "verbose", 'v', POPT_ARG_NONE, &Verbose, 0,
                  "show what is happening", "" },

                { "skip-already-indexed-check", 'S', POPT_ARG_NONE, &DontCheckIfAlreadyThere, 0,
                  "don't check if the context is already indexed, just add it.", "" },
                
                { "index-path", 'P', POPT_ARG_STRING, &IndexPath_CSTR, 0,
                  "which index to use", "" },

                { "filelist-file", 'f', POPT_ARG_STRING, &FilelistFile_CSTR, 0,
                  "file containing the URLs of the files to index (eg. made by find . >foo)", "" },

                { "filelist-stdin", '1', POPT_ARG_NONE, &FilelistStdin, 0,
                  "read filenames from stdin to index", "" },
                
                { "ignore-ea", 0, POPT_ARG_STRING, &IgnoreEANames, 0,
                  "ignore the following attributes when creating the ea index", "" },

                { "ignore-ea-append", 0, POPT_ARG_STRING, &IgnoreEANamesAppend, 0,
                  "append the following to the list of ea names to ignore when creating index",
                  "" },

                { "ignore-ea-regex-append", 0, POPT_ARG_STRING, &IgnoreEARegexsAppend, 0,
                  "append the following to the list of regexes which select which attributes are not to be indexed",
                  "" },

                { "ea-value-max-size", 0, POPT_ARG_INT, &EAIndexMaxValueSize, 0,
                  "largest attribute value to index", "" },

                { "ea-to-index", 0, POPT_ARG_STRING, &EAToIndex, 0,
                  "index only the attributes in the comma seperated list. good for automated testing. Note that this option makes --ignore-ea-regex-append, --ignore-ea-append and --ignore-ea impotent", "" },
                
                { "directory", 'd', POPT_ARG_NONE,
                  &ListDirectoryNodeOnly, 0,
                  "(like ls -d) index directory entries instead of contents", 0 },

                { "force-read-root-dir-nodes", 0, POPT_ARG_NONE,
                  &ForceReadRootDirectoryNodes, 0,
                  "Always read the context given on cmd line."
                  "Handy in use with -d to force EA generation", 0 },

                { "recursive", 'R', POPT_ARG_NONE, &RecursiveList, 0,
                  "index directorys recursively", 0 },

                { "ferris-filter", 0, POPT_ARG_STRING, &FilterStringCSTR, 0,
                  "Only index contexts which pass given filter",
                  "(name=fred*)" },

                { "show-totals", 0, POPT_ARG_NONE, &ShowTotals, 0,
                  "Display statistics at the end for number of files indexed etc.", 0 },

                { "force", 'F', POPT_ARG_NONE,
                  &AddEvenIfAlreadyCurrent, 0,
                  "Add all files again even if they have not changed between indexing", 0 },
                
//                 { "create-type", 0, POPT_ARG_STRING, &CreateTypeName_CSTR, 0,
//                   "what form of context to store the chunks in (dir/db4/gdbm/xml/etc)", 0 },

                { "retire-non-existent-to-multiversion", 'K', POPT_ARG_NONE,
                  &RetireNonExistentDocumentsToMuliversion, 0,
                  "Move metadata about files which no longer exist into multiversion instance now.", 0 },
                
                FERRIS_POPT_OPTIONS
                POPT_AUTOHELP
                POPT_TABLEEND
            };
        poptContext optCon;

        optCon = poptGetContext(PROGRAM_NAME.c_str(), argc, (const char**)argv, optionsTable, 0);
        poptSetOtherOptionHelp(optCon, "[OPTIONS]* src1 src2 ...");

        /* Now do options processing */
        char c=-1;
        while ((c = poptGetNextOpt(optCon)) >= 0)
        {}

        if (argc < 2 )
        {
            poptPrintHelp(optCon, stderr, 0);
            exit(1);
        }

        EAIndex::fh_idx idx;

        if( IndexPath_CSTR )
        {
            idx = EAIndex::Factory::getEAIndex( IndexPath_CSTR );
        }
        else
        {
            idx = EAIndex::Factory::getDefaultEAIndex();
        }

        string FilterString  = FilterStringCSTR ? FilterStringCSTR : "";
        fh_eaindexadd_display d = new Ferrisls_feaindexadd_display( idx );
        
        ls.setListDirectoryNodeOnly( ListDirectoryNodeOnly );
        ls.setForceReadRootDirectoryNodes( ForceReadRootDirectoryNodes );
        ls.setRecursiveList( RecursiveList );
        ls.setDisplay( d );
        
        if( FilterString.length() )
        {
            ls.setFilterString( FilterString );
        }

        if( IgnoreEANames )
            d->setIgnoreEANames( IgnoreEANames );
        if( IgnoreEANamesAppend )
            d->appendIgnoreEANames( IgnoreEANamesAppend );
        if( IgnoreEARegexsAppend )
            d->appendIgnoreEARegexs( IgnoreEARegexsAppend );
        d->setMaximunValueSize( EAIndexMaxValueSize );
        d->setDontCheckIfAlreadyThere( DontCheckIfAlreadyThere );
        d->setShowTotals( ShowTotals );
        d->setAddEvenIfAlreadyCurrent( AddEvenIfAlreadyCurrent );
        d->setRetireNonExistentDocumentsToMuliversion( RetireNonExistentDocumentsToMuliversion );
        
        if( EAToIndex )
        {
            stringlist_t sl = Util::parseCommaSeperatedList( EAToIndex );
            d->setEANamesToIndex( sl );
        }
        d->setVerbose( Verbose );

        m_totalRuntime.start();
        
        if( FilelistFile_CSTR )
        {
            string filelistFile = FilelistFile_CSTR;
            fh_ifstream fiss( filelistFile );
            addToIndexFromFileList( idx, fiss );
        }
        if( FilelistStdin )
        {
            fh_istream fiss = Ferris::Factory::fcin();
            addToIndexFromFileList( idx, fiss );
        }
        
        while( const char* tmpCSTR = poptGetArg(optCon) )
        {
            string srcURL = tmpCSTR;
            try
            {
                if( Verbose )
                    cerr << "srcurl:" << srcURL << endl;

                ls.setURL( srcURL );
                ls();
                d->DetachAllSignals();
            }
            catch( exception& e )
            {
                cerr << "----------------------" << endl;
                cerr << "for: " << srcURL << endl;
                cerr << "cought error:" << e.what() << endl;
                exit_status = 1;
            }
        }
//        d->sync();
        d->printTotals();

        TotalFilesDoneCount += d->getFilesIndexedCount();

        m_totalRuntime.stop();
        cerr << "Total contexts indexed:" << TotalFilesDoneCount << endl;
        m_totalRuntime.print();

        if( TotalFilesDoneCount )
        {
            double avgTimePerContext = m_totalRuntime.getElapsedTime() / TotalFilesDoneCount;
            cerr << "Average time per context:" << setprecision(3) << avgTimePerContext << endl;
        }
        
        
//         while( const char* tmpCSTR = poptGetArg(optCon) )
//         {
//             try
//             {
//                 string srcURL = tmpCSTR;
//                 cerr << "srcurl:" << srcURL << endl;
//                 fh_context c  = Resolve( srcURL );

//                 EAIndex::fh_docindexer indexer = EAIndex::Factory::makeDocumentIndexer( idx );
//                 if( IgnoreEANames )
//                     indexer->setIgnoreEANames( IgnoreEANames );
//                 if( IgnoreEANamesAppend )
//                     indexer->appendIgnoreEANames( IgnoreEANamesAppend );
//                 if( IgnoreEARegexsAppend )
//                     indexer->appendIgnoreEARegexs( IgnoreEARegexsAppend );
//                 indexer->setMaximunValueSize( EAIndexMaxValueSize );
//                 indexer->setDontCheckIfAlreadyThere( DontCheckIfAlreadyThere );

//                 if( EAToIndex )
//                 {
//                     stringlist_t sl = Util::parseCommaSeperatedList( EAToIndex );
//                     indexer->clearIgnoreEARegexs();
//                     indexer->clearIgnoreEANames();
//                     indexer->setEANamesToIndex( sl );
//                 }
                
                    
// //                     if( Verbose )
// //                         indexer->getProgressSig().connect( slot( progressf ) );
//                 indexer->addContextToIndex( c );
//             }
//             catch( exception& e )
//             {
//                 cerr << "cought error:" << e.what() << endl;
//                 exit_status = 1;
//             }
//         }
    }
    catch( exception& e )
    {
        cerr << "cought error:" << e.what() << endl;
        exit(1);
    }
    
    return exit_status || ls.hadErrors();
}


