/******************************************************************************
*******************************************************************************
*******************************************************************************

    ferris internal metadata worker direct client
    Copyright (C) 2007 Ben Martin

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

    For more details see the COPYING file in the root directory of this
    distribution.

    $Id: ferris-internal-metadata-broker-direct-client.cpp,v 1.3 2007/08/31 21:30:12 ben Exp $

*******************************************************************************
*******************************************************************************
******************************************************************************/

#include <popt.h>
#include <signal.h>

#include <iostream>
#include <string>
#include <list>

typedef std::list<std::string> stringlist_t;

#include <dbus-c++/dbus.h>
#include <dbus-c++/glib-integration.h>
#include "DBusGlue/ferris-internal-metadata-broker-client-glue.h"

#include "Ferris/MetadataServer_private.hh"

using namespace std;

const string PROGRAM_NAME = "ferris-internal-metadata-worker-direct-client";

void usage(poptContext optCon, int exitcode, char *error, char *addl)
{
    
    poptPrintUsage(optCon, stderr, 0);
    if (error) fprintf(stderr, "%s: %s0", error, addl);
    exit(exitcode);
}

///////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////

static const char* DBUS_SERVER_NAME = "net.sf.witme.Libferris.Metadata.Broker";
static const char* DBUS_SERVER_PATH = "/net/sf/witme/libferris/Metadata";

int RETURN_CODE = 0;

bool running = true;
DBus::Glib::BusDispatcher dispatcher;

namespace Ferris
{
    class MetadataBroker
        :
        public net::sf::witme::Libferris::Metadata::Broker,
        public DBus::IntrospectableProxy,
        public DBus::ObjectProxy
    {
    public:

        MetadataBroker( DBus::Connection& connection, const char* path, const char* name )
            :
            DBus::ObjectProxy(connection, path, name)
            {
            }

        virtual void asyncGetResult( const ::DBus::Int32& reqid,
                                     const ::DBus::String& earl,
                                     const ::DBus::String& rdn,
                                     const std::vector< ::DBus::Byte >& value )
            {
                cerr << "asyncGetResult() req:" << reqid
                     << " earl:" << earl
                     << " rdn:" << rdn
                     << " value:" << value
                     << endl;
                running = false;
//                dispatcher.leave();
            }

        virtual void asyncGetFailed( const ::DBus::Int32& reqid,
                                     const ::DBus::String& earl,
                                     const ::DBus::Int32& error,
                                     const ::DBus::String& ename,
                                     const ::DBus::String& edesc )
            {
                cerr << "asyncGetFailed() req:" << reqid
                     << " earl:" << earl
                     << " error:" << error
                     << endl;
                running = false;
//                dispatcher.leave();
            }
        
        virtual void asyncPutCommitted( const ::DBus::Int32& reqid,
                                        const ::DBus::String& earl,
                                        const ::DBus::String& rdn )
            {
            }
        
        virtual void asyncPutFailed( const ::DBus::Int32& reqid,
                                     const ::DBus::String& earl,
                                     const ::DBus::Int32& errno,
                                     const ::DBus::String& ename,
                                     const ::DBus::String& edesc )
            {
            }
        
        
    };
};



////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////


void niam( int sig )
{
    running = false;
//	dispatcher.leave();
}


int main( int argc, const char** argv )
{
	signal(SIGTERM, niam);
	signal(SIGINT, niam);

	DBus::default_dispatcher = &dispatcher;

    unsigned long ShowVersion        = 0;
    unsigned long UseSyncAPI         = 0;
    const char*   Metadata_CSTR      = 0;

    struct poptOption optionsTable[] = {

        { "version", 0, POPT_ARG_NONE, &ShowVersion, 0,
          "show version information and quit", 0 },

        { "metadata", 'a', POPT_ARG_STRING, &Metadata_CSTR, 0,
          "metadata to extract", 0 },

        { "sync", 's', POPT_ARG_NONE, &UseSyncAPI, 0,
          "use the sync API to read the value", 0 },
        
        /*
         * Standard Ferris options
         */
//        FERRIS_POPT_OPTIONS
        POPT_AUTOHELP
        POPT_TABLEEND
    };
    poptContext optCon;


    optCon = poptGetContext(PROGRAM_NAME.c_str(), argc, argv, optionsTable, 0);
    poptSetOtherOptionHelp(optCon, "[OPTIONS]* url1 url2 ...");

    /* Now do options processing, get portname */
    char c=-1;
    while ((c = poptGetNextOpt(optCon)) >= 0)
    {}

    if (argc < 1 || !Metadata_CSTR )
    {
        cerr << "Arguments incorrect!" << endl;
//        poptPrintHelp(optCon, stderr, 0);
        poptPrintUsage(optCon, stderr, 0);
        exit(1);
    }
    
    if( ShowVersion )
    {
        cout << "ferris-internal-metadata-worker" << endl
             << "version: $Id: ferris-internal-metadata-broker-direct-client.cpp,v 1.3 2007/08/31 21:30:12 ben Exp $\n"
             << "Written by Ben Martin, aka monkeyiq" << endl
             << endl
             << "Copyright (C) 2007 Ben Martin" << endl
             << "This is free software; see the source for copying conditions.  There is NO\n"
             << "warranty; not even for MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE."
             << endl;
        exit(0);
    }


    stringlist_t srcs;
    while( const char* RootNameCSTR = poptGetArg(optCon) )
    {
        string RootName = RootNameCSTR;
        srcs.push_back( RootName );
    }
    
    string rdn = Metadata_CSTR;
    
    try
    {
        
        DBus::Connection conn = DBus::Connection::SessionBus();

        Ferris::MetadataBroker client(conn, DBUS_SERVER_PATH, DBUS_SERVER_NAME);

        stringlist_t::iterator srcsiter = srcs.begin();
        stringlist_t::iterator srcsend  = srcs.end();
        for( ; srcsiter != srcsend ; ++srcsiter )
        {
            string earl = *srcsiter;

            if( UseSyncAPI )
            {
                cout << Ferris::syncMetadataServerGet( earl, rdn ) << endl;
            }
            else
            {
//            cout << client.get( earl, rdn ) << endl;
                cout << client.asyncGet( earl, rdn ) << endl;
            }
        }

        if( !UseSyncAPI )
        {
            while( running )
                g_main_iteration( true );
            
//            dispatcher.enter();
        }
    }
    catch( exception& e )
    {
        cerr << "cought:" << e.what() << endl;
        exit(1);
    }
        
    poptFreeContext(optCon);
    cout << flush;
    return RETURN_CODE;
}
