/******************************************************************************
*******************************************************************************
*******************************************************************************

    convert-myrdf-1172-to-1173
    Copyright (C) 2005 Ben Martin

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

    For more details see the COPYING file in the root directory of this
    distribution.

    $Id: convert-myrdf-1172-to-1173.cpp,v 1.1 2006/12/07 07:04:54 ben Exp $

*******************************************************************************
*******************************************************************************
******************************************************************************/

/*
 * return 0 for success
 * return 1 for generic error
 * 
*/


#include <Ferris.hh>
#include <Redland_private.hh>
#include <popt.h>

using namespace std;
using namespace Ferris;
using namespace Ferris::RDF;
using namespace Ferris::redlandea;

const string PROGRAM_NAME = "convert-myrdf-1172-to-1173";

void usage(poptContext optCon, int exitcode, char *error, char *addl)
{
    poptPrintUsage(optCon, stderr, 0);
    if (error) fprintf(stderr, "%s: %s0", error, addl);
    exit(exitcode);
}

/********************************************************************************/
/********************************************************************************/
/********************************************************************************/
/********************************************************************************/
/********************************************************************************/
/********************************************************************************/

unsigned long Verbose              = 0;
//const char* oldURL_CSTR = 0;

int main( int argc, char** argv )
{
    int exit_status = 0;
    
    try
    {

        struct poptOption optionsTable[] =
            {
                { "verbose", 'v', POPT_ARG_NONE, &Verbose, 0,
                  "show what is happening", "" },

                FERRIS_POPT_OPTIONS
                POPT_AUTOHELP
                POPT_TABLEEND
            };
        poptContext optCon;

        optCon = poptGetContext(PROGRAM_NAME.c_str(), argc, (const char**)argv, optionsTable, 0);
        poptSetOtherOptionHelp(optCon, "[OPTIONS]* oldURL newURL");

        /* Now do options processing */
        {
            char c=-1;
            while ((c = poptGetNextOpt(optCon)) >= 0)
            {}
        }
        

        if (argc < 1)
        {
            poptPrintHelp(optCon, stderr, 0);
            exit(1);
        }

        fh_model m = getDefaultFerrisModel();
        StatementIterator siter = m->begin();
        StatementIterator e     = m->end();

        typedef list< pair< fh_node, fh_node > > polist_t;
        typedef map< fh_node, polist_t, fh_node_compare > s_polist_t;
        s_polist_t s_polist;
        
        for( ; siter != e; ++siter )
        {
            fh_statement st = *siter;
            fh_node s = st->getSubject();
            fh_node p = st->getPredicate();
            fh_node o = st->getObject();

//             cerr << "Subject:" << s->toString()
//                  << " Pred:" << p->toString()
//                  << " Obj:" << o->toString()
//                  << endl;

            string subjstr = s->toString();
            
            if( !starts_with( subjstr, "_:" ) && contains( subjstr, ":" ) )
            {
                s_polist[ s ].push_back( make_pair( p, o ) );
            }
        }


        typedef map< fh_node, fh_node, fh_node_compare > nodeMap_t;
        nodeMap_t uuidNodes;
        nodeMap_t mdnodes;
        
        int SubjectsConverted = 0;
        int POConverted = 0;
        int POAttrConverted = 0;
        
        for( s_polist_t::iterator iter = s_polist.begin();
             iter != s_polist.end(); ++iter )
        {
            fh_node s = iter->first;
            fh_node uuidnode = uuidNodes[ s ];
            if( !uuidnode )
            {
                string uuid =  Util::makeUUID();
                uuidnode = Node::CreateURI( getPredicateURI( uuid ) );
                uuidNodes[ s ] = uuidnode;

                cerr << "Created uuidnode for url:" << s->toString()
                     << " uuidnode:" << uuidnode->toString()
                     << endl;
            }
            
            m->set( s, ::Ferris::redlandea::uuidPredNode(), uuidnode );

            fh_node outOfBandPred = redlandea::uuidOutOfBandPredNode();
            for( polist_t::iterator pliter = iter->second.begin();
                 pliter != iter->second.end(); ++pliter )
            {
                fh_node p = pliter->first;
                fh_node o = pliter->second;

                if( starts_with( p->getURI()->toString(), "http://witme.sf.net/libferris.web/rdf/ferris-attr/" ))
                {
                    fh_node mdnode = mdnodes[ uuidnode ];
                    if( !mdnode )
                    {
//                        mdnode = Node::CreateURI( getPredicateURI( Util::makeUUID() ) );
                        mdnode = Node::CreateBlankNode();
                        
                        m->set( uuidnode, outOfBandPred, mdnode );
                        cerr << "Created mdnode for url:" << s->toString()
                             << " uuid:" << uuidnode->toString()
                             << " mdnode:" << mdnode->toString()
                             << endl;
                        mdnodes[ uuidnode ] = mdnode;
                    }
                    
                    m->set( mdnode, p, o );
                    ++POAttrConverted;
                }
                else
                {
                    m->set( uuidnode, p, o );
                }
                m->erase( s, p, o );
                ++POConverted;
            }
            ++SubjectsConverted;
        }

        cerr << "Converted " << SubjectsConverted << " subjects" << endl;
        cerr << "Converted " << POAttrConverted << " old out-of-line EA" << endl;
        cerr << "Converted " << POConverted << " predicate-object pairs." << endl;
        m->sync();
    }
    catch( exception& e )
    {
        cerr << "error:" << e.what() << endl;
        exit(1);
    }
    return exit_status;
}


